using System;
using System.IO;
using System.Threading;
/// <summary>
///  ------------------------------------------------------
/// <h1>
/// AsyncCopy class     File: AsyncCopy.cs
/// </h1>
/// Description: Code demonstrates how to do asynchronous unbufferd IO
///     from within the .NET framework.
/// This is the Appendix of the Micrsoft Technical Report by Kukol and Gray
///    titled: Sequential File Programming Patterns and Performance with .NET
///    The code is downloadable from:
///    at URL http://research.microsoft.com/research/downloads/
///
/// The program creates the 1GB file C:\temp\source.dat and copies it to C:\temp\target.dat
///    using double buffering (four outstanding IOs.)
/// In the end it deletes the source and target files.
///
///  Usage is: AsyncCopy.exe
///
///  Version: 1.0,  November 2004
///
///  Authors: Peter Kukol, Jim Gray, Microsoft Corporation
///           PeterK@microsoft.com, Gray@Microsoft.com
///
///  (c) Microsoft Corporation. All rights reserved.
///
///</summary>
namespace AsyncCopy {
    class asyncCopy {
        //////////////////////////////////////////////////////////////////////////////
        ///  globals
        const int BUFFERS     = 4;                  // number of outstanding requests
        const int BUFFER_SIZE = 1 <<20;             // request size, 1MB
        public static FileStream source;            // source file stream
        public static FileStream target;            // target file stream
        public static long pending = 0;             // number of I/O's in flight
        public static long totalBytes = 0;          // total bytes to process
        public static long bytesRead = 0;           // bytes read so far
        public static long bytesWritten = 0;        // bytes written so far
        public static Object WriteCountMutex = new Object(); // mutex to protect count

        //////////////////////////////////////////////////////////////////////////////
        /// structure to hold IO request buffer and result.
        public class AsyncRequestState {            // data that tracks each async request
            public byte[]           Buffer;         // IO buffer to hold read/write data
            public AutoResetEvent   ReadLaunched;   // Event signals start of read
            public long             bufferOffset;   // buffer strides thru file BUFFERS*BUFFER_SIZE
            public IAsyncResult     ReadAsyncResult; // handle for read requests to EndRead() on.
            public AsyncRequestState(int i){        // constructor
                bufferOffset = i * BUFFER_SIZE;     // offset in file where buffer reads/writes
                ReadLaunched = new AutoResetEvent(false); // semaphore says reading (not writing)
                Buffer = new byte[BUFFER_SIZE];     // allocates the buffer
        }   }                                       // end AsyncRequestState declaration
        //////////////////////////////////////////////////////////////////////////////
        /// Array of buffers and async request results.
        public static AsyncRequestState[] request = new AsyncRequestState[BUFFERS];

        ////////////////////////////////////////////////////////////////////////////////
        // Asynchronous Callback completes writes and issues next read
        public static void WriteCompleteCallback(IAsyncResult ar) {
            lock (WriteCountMutex) {                                // protect the shared variables
                int i = Convert.ToInt32(ar.AsyncState);             // get request index
                target.EndWrite(ar);                                // mark the write complete
                bytesWritten += BUFFER_SIZE;                        // advance bytes written
                request[i].bufferOffset += BUFFERS * BUFFER_SIZE;   // stride to next slot
                if (request[i].bufferOffset < totalBytes)           // if not all read, issue next read
                {                                                   //  
                    source.Position = request[i].bufferOffset;      // issue read at that offset
                    request[i].ReadAsyncResult = source.BeginRead(request[i].Buffer, 0, BUFFER_SIZE, null, i);
                    request[i].ReadLaunched.Set();
                }                                                   // else if read point beyond end of file.
                else pending--;                                     // retire this buffer, do not ask for more IO
            }
        }

        //////////////////////////////////////////////////////////////////////////////
        ///main routine implements asynchronous File.Copy(@"C:\temp\source.dat", @"C:\temp\target.dat");
        static void Main(string[] args)
        {
            Console.WriteLine("AsyncCopy.exe Version 1.0. (c) Microsoft Corporation. All rights reserved.\n"
                            + @" Creates 1GB file C:\temp\source.dat, copies it to C:\temp\source.dat "
                            + "\n using 1MB asynchronous IO requests, prints speed, deletes files, and ends.");
            try                                     // create the source file (with no data in it
            {
                FileStream fs = new FileStream(@"C:\temp\source.dat", FileMode.OpenOrCreate);
                fs.SetLength(1 << 30);              // create a 1GB file.
                fs.Close();                         // close will cause zeros to be written to the file.
            }
            catch (Exception e)
            {
                Console.WriteLine(@"AsyncCopy.exe error creating file C:\temp\target.dat. \n**************\n {0}'", e);
                throw new FileNotFoundException(@"AsyncCopy.exe error creating fileC:\temp\source.dat", e);
            }

            ////////////////////////////////////////////////////////////////////////////
            // declare the callback to handle write completions.
            AsyncCallback writeCompleteCallback = new AsyncCallback(WriteCompleteCallback);

            ////////////////////////////////////////////////////////////////////////////
            /// start a request timer
            double elapsedTime, CPUtime, CPUutilization;
            IOexamples.Timer time = new IOexamples.Timer();
            time.timerReset();
            ////////////////////////////////////////////////////////////////////////////
            /// open source and destination
            try
            {
                source = new FileStream(@"C:\temp\source.dat", // open source file
                        FileMode.Open,              // for read
                        FileAccess.Read,            // will read the file
                        FileShare.Read,             // allow other readers
                        BUFFER_SIZE,                // buffer size
                        true);                      // async I/O  
                target = new FileStream(@"C:\temp\target.dat", // create target file
                        FileMode.Create,            // fault if it exists
                        FileAccess.Write,           // will write the file
                        FileShare.None,             // exclusive access
                        BUFFER_SIZE,                // buffer size
                        true);                      // async I/O 
            }
            catch (Exception e)
            {
                Console.WriteLine(@"AsyncCopy.exe could not open source or target files in C:\temp\*.dat."
                                    + "\n**************\n {0}'", e);
            }

            ////////////////////////////////////////////////////////////////////////////
            ///  files are open, initialize async request structures.
            totalBytes = source.Length;              // Size of source file
            // initialize the request structure to point to each request.
            for (int i = 0; i < BUFFERS ; i++)  request[i] = new AsyncRequestState(i);

            ////////////////////////////////////////////////////////////////////////////
            // launch initial async reads
            for (int i = 0; i < BUFFERS; i++) {     // no callback on reads.
                request[i].ReadAsyncResult = source.BeginRead(request[i].Buffer, 0, BUFFER_SIZE, null, i);
                request[i].ReadLaunched.Set();      // say that read is launched
                lock (WriteCountMutex) pending++;
            }
            ////////////////////////////////////////////////////////////////////////////
            // wait for the reads to complete in order, process buffer and then write it.
            for (int i = 0; (bytesRead < totalBytes ); i = (i+1) % BUFFERS) {
                request[i].ReadLaunched.WaitOne();  // wait for flag that says buffer is reading
                int bytes = source.EndRead(request[i].ReadAsyncResult); // wait for read complete
                bytesRead += bytes;                 // process the buffer <your code goes here>
                target.BeginWrite(request[i].Buffer, 0, bytes, writeCompleteCallback, i); // write it
            }                                       // end of reader loop

            ////////////////////////////////////////////////////////////////////////////
            // IOs started now wait for them to all complete
            while (pending > 0) Thread.Sleep(10);   // wait for all the writes to complete
            source.Close(); target.Close();         // close the files

            ////////////////////////////////////////////////////////////////////////////
            /// report the time and speed.
            elapsedTime = time.timerRead(out CPUtime, out CPUutilization);
            float fileSizeMB = totalBytes >> 20;

            Console.WriteLine("AsyncCopy of a {0:F0} MB file in: {1:F2} seconds @ {2:F2} MBps, \n\tCPU: {3:F2},\tUser: {4:F2},\tKernel {5:F2}."
                             , fileSizeMB, elapsedTime, fileSizeMB / elapsedTime, CPUtime, time.UserTime, time.KernelTime);
            ////////////////////////////////////////////////////////////////////////////
            /// cleanup
            File.Delete(@"C:\temp\source.dat");     // clean up in the end, discard the source and target files.
            File.Delete(@"C:\temp\target.dat");     // clean up in the end
        }                                           // end of AsyncCopy.
    }
}
