using System;
using System.IO;
namespace IOexamples
{
/// <summary>
///  ------------------------------------------------------
/// <h1>
/// CountLines class     File: CountLines.cs
/// </h1>
/// Description: Demonstrate different IO read styles
/// <ul>
///     <li>Byte-At-A-Time</li>
///     <li>Line-At-A-Time</li>
///     <li>Block-At-A Time</li>
/// </ul>
///
///  Version: 1.0,  November 2004
///
///  Author: Peter Kukol, Jim Gray, Microsoft Corporation
///        PeterK@microsoft.com, Gray@Microsoft.com
///
///  (c) Microsoft Corporation. All rights reserved.
/// 
///</summary>

    public class CountLines
    {
        /// <summary>
        /// Description: Demonstrate different IO read styles
        /// <ul>
        ///     <li>Byte-At-A-Time</li>
        ///     <li>Line-At-A-Time</li>
        ///     <li>Block-At-A Time</li>
        /// </ul>
        /// </summary>
        /// <value>Returns count of lines in the named file.</value>
        /// <param name="fileName">Path name of file to read.</param>
        /// <param name="option">one of {"Byte-At-A-Time", "Line-At-A-Time", "Block-At-A-Time" }, default is  "Block-At-A-Time" </param>
        /// <returns> long count of lines. </returns>
        /// <exception> writes to console and then passes back any exceptions with a diagnostic added.  </exception>
        public long countLines(string fileName, string option)
        {
            long count = 0;             // count of lines in file
            try
            {                           // catch any file exceptions
                switch (option)         // switch on file access type
                {
                    //--------------------------------------------------------------
                    // Read the file one byte at a time looking for newline chars.
                    case "Byte-At-A-Time": // using opens and closes the stream
                        using (FileStream fs = new FileStream(fileName, FileMode.Open))
                        {
                            int nextByte;
                            while ((nextByte = fs.ReadByte()) > 0)
                            { // at end of file, ReadByte() returns -1
                                if (nextByte == '\n') count++;
                            }
                        }
                        return count;
                    //--------------------------------------------------------------
                    // Read the file one line at a time counting the lines.
                    case "Line-At-A-Time":
                        using (StreamReader sr = new StreamReader(fileName))
                        {
                            while (sr.ReadLine() != null)
                            {
                                count++;
                            }
                        }
                        return count;
                    //--------------------------------------------------------------
                    // Read the file one 64KB block at a time and count the newline chars
                    case "Block-At-A-Time":
                    default:
                        using (FileStream fs = new FileStream(fileName, FileMode.Open))
                        {
                            int fileIOblockSize = 1 << 16;              // default is all IO in 64KB blocks
                            byte[] IObuff = new byte[fileIOblockSize];  // buffer to hold bytes
                            while (fs.Position < fs.Length)
                            {
                                int readCount = fs.Read(IObuff, 0, fileIOblockSize);
                                for (int i = 0; i < readCount; i++)
                                    if (IObuff[i] == '\n') count++;
                            }
                        }
                        return count;
                }
            }
            //------------------------------------------------------------------
            // control comes here if there is a problem opening or reading the file.
            catch (Exception e)
            {
                Console.WriteLine("Error reading file {0}. \n**************\n {1}'", fileName, e);
                throw new FileNotFoundException("Error reading file: "               + fileName, e);
            }
        }
    }
}
