using System;
using System.IO;
using System.Diagnostics;
namespace FragDisk
{
//////////////////////////////////////////////////////////////////////////////
/// <summary>
    ///  ------------------------------------------------------
    /// <h1>
    /// FragDisk class     File: FragDisk.cs
    /// <br>fragments a given volume (by creating and deleting many files 
    /// within a given range of sizes). 
    /// </h1>
    /// Used to create a fragmented disk for IO experments described in 
    /// MSR Technical Report by Kukol and Gray titled: 
    /// Sequential File Programming Patterns and Performance with .NET
    ///    The code is downloadable from:
    ///    at URL http://research.microsoft.com/research/downloads/
    ///
    /// This main program 
    ///     parses the parameters and gets the base directory path.
    ///     Builds a directory tree under that path deep enough so that
    ///         the leaf directories have at most MaxFilesPerDirectory files
    ///         the non-leaf directories have at most MaxDirectoriesPerPirectory subdirectories.
    ///     It populates the leaf directories with files of size MinFileSize to MaxFileSize.
    ///     Filling the disk up to fillPercentage full. 
    ///     Then it deletest all but keepPercentage of these new files. 
    ///     That ends a grow-shrink cycle.  It executres -n of these cycles. 
    ///
    /// Usage is: FragDisk [options] directoryPath
    /// Options:
    /// -m<count>   set max. number of files to create               (default=100000)
    /// -Fm<size>   set min. file size (in MB)                       (default=1)
    /// -FM<size>   set max. file size (in MB)                       (default=256)
    /// -c<count>   set number of files per cycle                    (default=1000)
    /// -d<count>   set max. number of files per directory           (default=100)
    /// -s<count>   set max. number of sub-directories per directory (default=10)
    /// -n<count>   set max. create/delete cycles, 0 = no limit      (default=0)
    /// -k<pctg>    set percentage (1-99) of files to keep           (default=5)
    /// -f<pctg>    set desired percentage (1-99) to fill the volume (default=70)
    /// -r<value>   set random seed                                  (default=137)
    /// 
    ///  Version: 1.0,  November 2004
    ///
    ///  Authors: Peter Kukol, Jim Gray, Microsoft Corporation
    ///           PeterK@microsoft.com, Gray@Microsoft.com
    ///
    ///  (c) Microsoft Corporation. All rights reserved.
    /// 
    ///</summary>
    public class fragDisk
    {
        #region globals
        const int DEFAULT_MAX_FILES = 100000;         // maximum number of files
        const int DEFAULT_CYCLE_SIZE = 1000;           // maximum number of create/delete cycles
        const int DEFAULT_KEEP_PCTG = 5;              // default % of files to keep from each cycle
        const int DEFAULT_FILL_PCTG = 70;             // default % of disk to fill on each cycle.
        const int DEFAULT_FILES_PER_DIR = 100;        // default # files per directory
        const int DEFAULT_SDIRS_PER_DIR = 10;         // default # subdirectories per directory
        const int DEFAULT_MIN_FILE_SIZE = 1;          // in MB
        const int DEFAULT_MAX_FILE_SIZE = 256;        // in MB
        const int DEFAULT_CYCLE_CNT = 0;              // default is no cycle limit
        const int DEFAULT_RANDOM_SEED = 137;          // you can override this to make experiments repeatable
        // initialization
        static bool verbose = false;                    // flag controls volume of trace outupt
        static int maxFiles = DEFAULT_MAX_FILES;        // max files to create in a directory
        static double keepFract = DEFAULT_KEEP_PCTG / 100.0; // fraction of files to keep on each shrink cycle
        static double fillFract = DEFAULT_FILL_PCTG / 100.0; // fraction of disk to fill on each fill cycle
        static int filesPerCycle = DEFAULT_CYCLE_SIZE;  // number of files to create per cycle
        static int filesPerDir = DEFAULT_FILES_PER_DIR; // number of files to create per leaf directory
        static int sdirsPerDir = DEFAULT_SDIRS_PER_DIR; // sub-directories per non-leaf directory.
        static int minFileSize = DEFAULT_MIN_FILE_SIZE; // new file sizes are picked between min and max
        static int maxFileSize = DEFAULT_MAX_FILE_SIZE; //
        static int cycleCnt = DEFAULT_CYCLE_CNT;        // number of grow-shrink cycles. 
        static string basePath;                         // root directory of the fragmented files
        static int randomSeed = DEFAULT_RANDOM_SEED;    // set this to make experment repeatable. 
        static Random frand;                            // random number used in file generation.
        #endregion globals
        #region commandHandling
        /// <summary>
        /// Display usage string.
        /// </summary>
        static void usage()
        {
            Console.WriteLine("Usage: FragDisk [options] directory_path");
            Console.WriteLine("");
            Console.WriteLine("Options:");
            Console.WriteLine("");
            Console.WriteLine("-m<count>   set max. number of files to create               (default=" + DEFAULT_MAX_FILES + ")");
            Console.WriteLine("-Fm<size>   set min. file size (in MB)                       (default=" + DEFAULT_MIN_FILE_SIZE + ")");
            Console.WriteLine("-FM<size>   set max. file size (in MB)                       (default=" + DEFAULT_MAX_FILE_SIZE + ")");
            Console.WriteLine("-c<count>   set number of files per cycle                    (default=" + DEFAULT_CYCLE_SIZE + ")");
            Console.WriteLine("-d<count>   set max. number of files per directory           (default=" + DEFAULT_FILES_PER_DIR + ")");
            Console.WriteLine("-s<count>   set max. number of sub-directories per directory (default=" + DEFAULT_SDIRS_PER_DIR + ")");
            Console.WriteLine("-n<count>   set max. create/delete cycles, 0 = no limit      (default=" + DEFAULT_CYCLE_CNT + ")");
            Console.WriteLine("-k<pctg>    set percentage (1-99) of files to keep           (default=" + DEFAULT_KEEP_PCTG + ")");
            Console.WriteLine("-f<pctg>    set desired percentage (1-99) to fill the volume (default=" + DEFAULT_FILL_PCTG + ")");
            Console.WriteLine("-r<value>   set random seed                                  (default=" + DEFAULT_RANDOM_SEED + ")");
            Console.WriteLine("-v          run in verbose mode");
            Console.WriteLine("{0} version {1} fragments a disk based on specified parameters.",
                                System.Reflection.Assembly.GetExecutingAssembly().GetName().Name,
                                System.Reflection.Assembly.GetExecutingAssembly().GetName().Version.ToString());
            Console.WriteLine("Copyright (C) Microsoft Corporation 2004. All rights reserved.\n");
        }
        /// <summary>
        /// Convert "+" and "-" toggle options to boolean
        /// </summary>
        /// <param name="optStr">the "+" or "-" or null string (== "+") </param>
        /// <param name="optChar">the option being controlled</param>
        /// <returns>true if option "+", false if option "-", exit otherwise.</returns>
        static bool checkToggleOpt(string optStr, char optChar)
        {
            if (optStr.Equals("")) return true;
            if (optStr.Equals("+")) return true;
            if (optStr.Equals("-")) return false;

            Console.WriteLine("Invalid '" + optChar + "' option: use -" + optChar + " to enable, -" + optChar + "- to disable");
            usage();
            System.Environment.Exit(-1);
            return false;
        }
        /// <summary>
        /// parseCommand(string [] argv) parses the command string
        /// </summary>
        /// <param name="args">The command line argument array given to Main()</param>
        /// <returns>true if command string OK, false otherwise exit otherwise.</returns>
        static bool parseCommand(string[] args)
        {
            int argCnt = 0;             // counts the number of processed command options.
            // Loop over all the command options till one does not start with - or /
            foreach (string arg in args)
            {
                // Stop when the argument is not a option
                if (arg[0] != '-' && arg[0] != '/') break;
                // if not at least 2 long, then it is a syntax error
                if (arg.Length < 2)
                {
                    Console.WriteLine("Invalid option: '" + arg + "'");
                    goto BAD_ARGS;
                }
                string argVal = arg.Substring(2); // get flag suffix if it exists.

                switch (arg[1])
                {
                    /////////////////////////////////////////////////
                    // Number of create-delete cycles.
                    case 'c':
                        try
                        {
                            filesPerCycle = Convert.ToInt32(argVal);
                            if (filesPerCycle > 0) break;
                        }
                        catch (Exception e)
                        {
                            Console.WriteLine(e.ToString());
                        }
                        Console.WriteLine("Invalid cycle count argument: '" + arg + "'");
                        goto BAD_ARGS;
                    /////////////////////////////////////////////////
                    // Max number of files per directory.
                    case 'd':
                        try
                        {
                            filesPerDir = Convert.ToInt32(argVal);
                            if (filesPerDir > 0) break;
                        }
                        catch (Exception e)
                        {
                            Console.WriteLine(e.ToString());
                        }
                        Console.WriteLine("Invalid max. files/directory argument: '" + arg + "'");
                        goto BAD_ARGS;
                    /////////////////////////////////////////////////
                    // Percentage of disk that will be filled at each cycle.
                    case 'f':
                        try
                        {
                            int pctg = Convert.ToInt32(argVal);
                            if (pctg > 1 && pctg < 100)
                            {
                                fillFract = pctg / 100.0;
                                break;
                            }
                        }
                        catch (Exception e)
                        {
                            Console.WriteLine(e.ToString());
                        }
                        Console.WriteLine("Invalid disk fullness percentage argument: '" + arg + "'");
                        goto BAD_ARGS;

                    /////////////////////////////////////////////////
                    // Min/max file size.
                    case 'F':
                        if (argVal.Length > 0)
                        {
                            bool ismax;
                            try
                            {
                                if (arg[2] == 'm')
                                    ismax = false;
                                else if (arg[2] == 'M')
                                    ismax = true;
                                else
                                    goto BAD_FSIZ;
                                int size = Convert.ToInt32(argVal.Substring(1));
                                if (size > 0)
                                {
                                    if (ismax)
                                        maxFileSize = size;
                                    else
                                        minFileSize = size;
                                    break;
                                }
                            }
                            catch (Exception e)
                            {
                                Console.WriteLine(e.ToString());
                            }
                        }
                    BAD_FSIZ:
                        Console.WriteLine("Invalid file size argument: '" + arg + "'");
                        goto BAD_ARGS;

                    /////////////////////////////////////////////////
                    // Percentage of files that survive each cycle.
                    case 'k':
                        try
                        {
                            int pctg = Convert.ToInt32(argVal);
                            if (pctg > 1 && pctg < 100)
                            {
                                keepFract = pctg / 100.0;
                                break;
                            }
                        }
                        catch (Exception e)
                        {
                            Console.WriteLine(e.ToString());
                        }
                        Console.WriteLine("Invalid files-to-keep percentage argument: '" + arg + "'");
                        goto BAD_ARGS;
                    /////////////////////////////////////////////////
                    // Maximum files to create.
                    case 'm':
                        try
                        {
                            maxFiles = Convert.ToInt32(argVal);
                            if (maxFiles > 0) break;
                        }
                        catch (Exception e)
                        {
                            Console.WriteLine(e.ToString());
                        }
                        Console.WriteLine("Invalid max. file count argument: '" + arg + "'");
                        goto BAD_ARGS;
                    /////////////////////////////////////////////////
                    // Maximum files to create.
                    case 'n':
                        try
                        {
                            cycleCnt = Convert.ToInt32(argVal);
                            if (cycleCnt > 0) break;
                        }
                        catch (Exception e)
                        {
                            Console.WriteLine(e.ToString());
                        }
                        Console.WriteLine("Invalid max. cycle count argument: '" + arg + "'");
                        goto BAD_ARGS;
                    /////////////////////////////////////////////////
                    // random seed control so experiments are repeatable.
                    case 'r':
                        try
                        {
                            randomSeed = Convert.ToInt32(argVal);
                            if (randomSeed > 0) break;
                        }
                        catch (Exception e)
                        {
                            Console.WriteLine(e.ToString());
                        }
                        Console.WriteLine("Invalid random seed argument: '" + arg + "'");
                        goto BAD_ARGS;
                    /////////////////////////////////////////////////
                    // Sub-directories per directory
                    case 's':
                        try
                        {
                            sdirsPerDir = Convert.ToInt32(argVal);
                            if (sdirsPerDir > 0) break;
                        }
                        catch (Exception e)
                        {
                            Console.WriteLine(e.ToString());
                        }
                        Console.WriteLine("Invalid max. sub-directories argument: '" + arg + "'");
                        goto BAD_ARGS;
                    /////////////////////////////////////////////////
                    // verbose
                    case 'v':
                        verbose = checkToggleOpt(argVal, 'v');
                        break;
                    /////////////////////////////////////////////////
                    // display usage
                    case '?':
                        // usage(); parent will do this
                        goto BAD_ARGS;
                    /////////////////////////////////////////////////
                    // unknown switch type.
                    default:
                        Console.WriteLine("Invalid option: '" + arg + "'");
                        goto BAD_ARGS;
                }
                // processed a good argument, increment the count.
                argCnt++;
            }
            //////////////////////////////////////////////////////////////////////////////////////////
            // After all parameters are parsed, there should be a single directory/drive argument
            if (argCnt != args.Length - 1)
            {
                Console.WriteLine("Expected a single drive/directory path argument, found " + argCnt + " argument(s) instead.");
                goto BAD_ARGS;
            }
            //////////////////////////////////////////////////////////////////////////
            // Everything was OK.
            return true;

            //////////////////////////////////////////////////////////////////////////
            // come here when things are bad.
            BAD_ARGS:
            usage();
            return false;
        }
        #endregion commandHandling
        #region Main
        //////////////////////////////////////////////////////////////////////////
        //  Main routine:  Parse command, then launch fragmenter.
        //////////////////////////////////////////////////////////////////////////
        public static void Main(String[] args)
        {
            /////////////////////////////////////////////////////////////////////
            // parse the command string.
            if (!parseCommand(args)) return;
            Console.WriteLine("{0} version {1}."
                                 + " (c) Microsoft Corporation. All rights reserved.\n",
                                    System.Reflection.Assembly.GetExecutingAssembly().GetName().Name,
                                    System.Reflection.Assembly.GetExecutingAssembly().GetName().Version.ToString());
            /////////////////////////////////////////////////////////////////////
            // Initialize the random number generator.
            frand = new Random(randomSeed);
            /////////////////////////////////////////////////////////////////////
            // set the base directory path (parseCommand() above checks we have it)
            basePath = args[args.Length - 1];
            /////////////////////////////////////////////////////////////////////
            // fragment the disk.
            frag();
        }
        #endregion Main
        #region FileDesc
        ///////////////////////////////////////////////////////////////////////////////
        // We maintain a "shadow" directory structure for our base directory tree
        // with an entry for each directory with its list of files and list of subdirectories
        // each file also has a descriptor.
        // The "growth" cycle grows this tree, the "shrink" cycle deletes random files from the tree.
        ///////////////////////////////////////////////////////////////////////////////
        public class fileDesc
        {
            public fileDesc fdParent;       // parent dir or null if top-level
            public bool fdIsFile;       // flag says this is a file descriptor (rather than a directory)
            public int fdIndex;        // negative if this is a directory
            public string fdPath;         // full disk path to this file/dir

            public fileDesc[] fdFileTab;      // files in this sub-directory
            public int fdFileCnt;
            public int fdFileMax;      // highest file index used

            public fileDesc fdFileFree;     // list of free file entries
            public fileDesc fdFileNext;     // next    free file entry

            public fileDesc[] fdSDirTab;      // sub-directories
            public int fdSDirCnt;      // number of subdirectories
            ///////////////////////////////////////////////////////////////////////////////
            //  empty constructor
            public fileDesc()
            {
            }
            ///////////////////////////////////////////////////////////////////////////////
            // initialize a file/directory descriptor.
            ///////////////////////////////////////////////////////////////////////////////
            public void init(bool isDir, int index, string path, fileDesc parent)
            {
                Debug.Assert(index > 0);
                if (!isDir)
                {
                    Debug.Assert(parent != null);
                    if (parent.fdFileMax < index)
                        parent.fdFileMax = index;
                }
                // at first diretory has no sub directories or files.
                this.fdSDirCnt =
                this.fdFileCnt =
                this.fdFileMax =
                this.fdSDirCnt = 0;
                // the lists are empty
                this.fdSDirTab =
                this.fdFileTab = null;
                //???
                this.fdIsFile = !isDir;
                this.fdIndex = index;
                this.fdParent = parent;
                this.fdPath = (path == null) ? fragPath(this) : path;
            }
        };
        #endregion FileDesc
        #region frag
        /// <summary>
        /// This is the procedure that drives the overall fragmentation process.
        ///
        /// First we find any pre-existing files / sub-directories within the
        /// target directory path, and we also make sure that we have enough
        /// sub-directories for the needed number of files (we don't want to put
        /// too many files in any given directory because this could slow things
        /// down unnecessarily).
        ///
        /// In order to fragment the disk, we repeatedly create and then delete
        /// bunches of files in a random fashion, taking care to only fill the
        /// drive to the 'fullness' level indicated by the user. The size of
        /// files varies randomly between the desired min and max sizes, with
        /// small files being created a lot more often than larger file (which
        /// causes the drive to become fragmented faster).
        /// The growth phase ends when the disk goes above fillPctg percent full.
        /// The shrink phase ends when the frag has deleted newcnt * (1 - keepFract) files.
        /// It executes cycleCnt such cycles.. 
        /// Frag execu
        /// </summary>
        static void frag()
        {
            int ldirCnt;
            fileDesc[] leafDirs;

            string drvStr;
            bool shrink;
            double fullness;               // indicates how full the drive is

            double sizeSpanMax = (maxFileSize - minFileSize) * 1024 * 1024;
            double sizeSpanMin = (maxFileSize - minFileSize) * 1024 * 10;

            leafDirs = new fileDesc[maxFiles / filesPerDir + 1];
            createDirTable(leafDirs, out ldirCnt);

            /////////////////////////////////////////////////////////////////
            // Get the target drive letter and see how full the drive is
            drvStr = Path.GetFullPath(basePath).Substring(0, 2);
            fullness = diskFullness(drvStr);

            /////////////////////////////////////////////////////////////////
            // If the drive is already full enough, start shrinking by deleting
            // files; otherwise start creating new files.
            shrink = (fullness > fillFract);

            /////////////////////////////////////////////////////////////////
            // Ready to start filling the disk
            while (true)
            {
                int newcnt = 0;                         // count of files in this cycle.
                //////////////////////////////////////////////////////////////////
                // Create a bunch of files (unless the disk is already too full)
                if (!shrink)
                {
                    /* Create up to filesPerCycle files for this cycle */
                    for (int fileNumber = 0; fileNumber < filesPerCycle; fileNumber++)
                    {
                        fileDesc fdesc = null;
                        // try to create a file in a random directory
                        for (int tries = 0; tries < ldirCnt; tries++)
                        {
                            /* Pick a random leaf directory */
                            fileDesc ddesc = leafDirs[(int)(ldirCnt * frand.NextDouble())];
                            Debug.Assert(ddesc != null);
                            /* Are there any free file entries in this directory? */
                            fdesc = ddesc.fdFileFree;
                            if (fdesc == null) continue;
                            /* Remove the file entry from the free list */
                            ddesc.fdFileFree = fdesc.fdFileNext;
                            /* Create a file descriptor */
                            fdesc.init(false, ++ddesc.fdFileMax, null, ddesc);
                            break;
                        }
                        Debug.Assert(fdesc.fdIsFile);
                        /* Create the actual physical file */
                        FileStream stream = File.Create(fdesc.fdPath);
                        newcnt++;
                        if (verbose) Console.WriteLine("Creating file '" + fdesc.fdPath + "'");
                        ////////////////////////////////////////////////////////////////
                        //  Extend the file to a random size (within the specified range,
                        //   with smaller files being 10x more numerous).
                        double sspan = (frand.NextDouble() <= 0.1) ? sizeSpanMin : sizeSpanMax;
                        long fsize = (long)(minFileSize + frand.NextDouble() * sspan);
                        try { stream.SetLength(fsize); }
                        catch (System.IO.IOException)
                        {
                            /* The disk is presumably full [ISSUE: How to detect this for sure?] */
                            if (verbose)
                                Console.WriteLine("NOTE: Looks like the disk is full, done creating files (disk fullness = "
                                                + (int)(100 * diskFullness(drvStr) + 0.5) + ")");
                            stream.Close();
                            break;
                        }
                        stream.Close();
                    }   // bottom of the creating loop, all the files created.
                }

                //////////////////////////////////////////////////////////////////////
                // Now enter the shrink cycle.
                // Delete proper percentage of files at random from the directory tree.
                int delcnt = Math.Max(10, (int)(newcnt * (1 - keepFract)));
                while (delcnt > 0)
                {
                    fileDesc ddesc = null;      // descriptor of a random leaf directory
                    fileDesc fdesc = null;      // descriptor of a random file directory
                    while (true)
                    {   /////////////////////////////////////////////////////////////
                        /// Pick a random leaf directory  
                        int ldir = (int)(ldirCnt * frand.NextDouble());
                        ddesc = leafDirs[ldir]; Debug.Assert(ddesc != null);
                        /////////////////////////////////////////////////////////////
                        /// Pick a random file entry and see if the file exists  
                        int filex = (int)(ddesc.fdFileCnt * frand.NextDouble()); // pick random file index
                        fdesc = ddesc.fdFileTab[filex];     // get it's descriptor.
                        if (fdesc.fdIndex != 0)             // if no such file guess again.
                            break;
                    }
                    ////////////////////////////////////////////////////////////////
                    /// found a random file, now delete it
                    Debug.Assert(fdesc != null); Debug.Assert(ddesc == fdesc.fdParent);
                    if (verbose) Console.WriteLine("Removing file '" + fdesc.fdPath + "'");
                    ////////////////////////////////////////////////////////////////
                    /// do the delete
                    File.Delete(fdesc.fdPath);
                    ////////////////////////////////////////////////////////////////
                    /// clear its index entry in the directory descriptor
                    fdesc.fdIndex = 0;
                    fdesc.fdFileNext = ddesc.fdFileFree;
                    ddesc.fdFileFree = fdesc;
                    ////////////////////////////////////////////////////////////////
                    /// increment count of deleted files.
                    delcnt--;
                }
                //////////////////////////////////////////////////////////////////////////////
                // Is the disk full enough to end this cycle?
                fullness = diskFullness(drvStr);
                if (verbose)
                    Console.WriteLine("A cycle is complete, disk fullness is at " + (int)(fullness * 100) + "%");

                ////////////////////////////////////////////////////////////////
                /// see if we are at the end of a grow+shrink cycle.
                if (shrink)
                {   // shrink cycle ends when fullness goes below fillFract
                    if (fullness <= fillFract) break;
                }
                else
                {   // growth cycle ends when fullness goes above fillFract
                    if (fullness >= fillFract) break;
                }
                // Check for cycle count limit
                if (cycleCnt > 0)
                    if (--cycleCnt == 0) break;
            }       // end loop on grow-shrink cycles.
        }           // end fragmentation program.
        #endregion frag
        #region FragPath
        ///////////////////////////////////////////////////////////////////////////////
        /// <summary>
        /// Given a file or directory descriptor, returns its full full path name.
        /// </summary>
        /// <param name="fdesc">Directory descriptor</param>
        /// <returns>Full path to directory</returns>
        ///////////////////////////////////////////////////////////////////////////////
        static string fragPath(fileDesc fdesc)
        {
            Debug.Assert(fdesc != null);
            int index = fdesc.fdIndex;
            char tchar = (fdesc.fdIsFile ? 'F' : 'D');
            string prefix = (fdesc.fdParent == null) ? basePath : fragPath(fdesc.fdParent);
            return prefix + "\\" + tchar + index.ToString("0000");
        }
        #endregion FragPath
        #region createDirTable
        /// <summary>
        /// Create the directory hierarchy underneath the root path that is large enough to host maxFiles files.
        /// Calls findDirsAndFiles() which recursively calls itself to do the actual work.
        /// </summary>
        /// <param name="leafDirs"></param>
        /// <param name="ldirCnt">Output param shows the number of leaf directories</param>
        static void createDirTable(fileDesc[] leafDirs, out int ldirCnt)
        {
            int dirCount = maxFiles / sdirsPerDir;
            int dirLevels;
            ldirCnt = 0;

            /////////////////////////////////////////////////////////////////////
            // compute how many levels are needed (filesPerDir^levels > maxFiles)
            long temp = filesPerDir;
            for (dirLevels = 1; temp < maxFiles; dirLevels++)
                temp = temp * sdirsPerDir;
            /////////////////////////////////////////////////////////////////////
            // The parent of the root directory will be null.
            fileDesc toss = null;
            /////////////////////////////////////////////////////////////////////
            // Recursively create the directory hierarchy
            int fcnt = findDirsAndFiles(maxFiles,       // file count
                0,              // directory index
                null,           // parent
                ref toss,       // this directory
                leafDirs,       // leaf directories
                ref ldirCnt,    // next leaf directory index
                1,              // current level
                dirLevels);     // max levels needed

            // Make sure we created the expected number of files
            Debug.Assert(fcnt == 0);
        }
        #endregion createDirTable
        #region findDirsAndFiles
        /// <summary>
        /// findDirsAndFiles() creates the shadow directory entry for some real directory. 
        ///    If the directory is exists, findDirsAndFiles builds the shadow entries 
        ///         from the real files and dirs. 
        ///    If the directory does not exist it creates the directory and its shadow.
        ///    If the directoy is not a leaf, it creates the needed number of sub-directories 
        ///         and calls findDirsAndFiles() on each one. 
        ///    If the directory is a leaf, it creates the requisite files in that directory. 
        /// </summary>
        /// <param name="fileCnt">Number of files left to find/create</param>
        /// <param name="dirIndex">Index of the current directory</param>
        /// <param name="parent">Parent directory descriptor</param>
        /// <param name="thisDir">Output parameter set to the directory created</param>
        /// <param name="leafDirs">The global table of leaf directories</param>
        /// <param name="ldirNum">Index of this directory within the leaf directory table</param>
        /// <param name="curLevel">Current level of directory nesting</param>
        /// <param name="maxLevel">Max levels of directory nesting we'll need</param>
        /// <returns>Number of files left to be created</returns>
        ///////////////////////////////////////////////////////////////////////////////
        static int findDirsAndFiles(         int fileCnt,
                                             int dirIndex,
                                             fileDesc parent,
                                         ref fileDesc thisDir,
                                             fileDesc[] leafDirs,
                                         ref int ldirNum,
                                             int curLevel,
                                             int maxLevel)
        {
            int filcnt,                     // target   file count
                filxc;                      // existing file count
            int dircnt,                     // target   subDirectory count
                dirxc;                      // existing subDirectory count
            int fndcnt;                     // for file/dir count verification only
            fileDesc[] fkids;               // descriptors for pre-existing files in this directory
            fileDesc[] fdirs;               // descriptors for pre-existing sub-directories in this directory.
            ///////////////////////////////////////////////////////////////////////////
            // Initialize current shadow directory entry if it is null
            if (thisDir == null) thisDir = new fileDesc();
            fileDesc ddesc = thisDir;
            ddesc.init(true, dirIndex, ddesc.fdPath, parent);
            /////////////////////////////////////////////////////////////////////
            // Create corresponding real directory if it doesn't exist, ohterwise count its contents
            if (Directory.Exists(ddesc.fdPath))
            {   // if directory already exists, count/enumerate its contents
                dirxc = enumDirectory(ddesc, true, null);  // count the sub-directories
                filxc = enumDirectory(ddesc, false, null);  // count the files
            }
            else
            {   // if it does not exist, create an empty directory (fills later)
                try
                {
                    Directory.CreateDirectory(ddesc.fdPath);
                }
                catch
                {
                    Console.WriteLine("Could not create directory '" + ddesc.fdPath + "', aborting.");
                    System.Environment.Exit(-1);
                }
                filxc = dirxc = 0;
            }
            /////////////////////////////////////////////////////////////////////
            // Make sure we create enough files and sub-dirs
            if (curLevel < maxLevel)
            {
                /// This is not yet the leaf level, create sub-Directories
                dircnt = Math.Max(dirxc, sdirsPerDir);
                filcnt = 0;                     // no files in non-leaf directories
            }
            else
            {
                /// This is the leaf level, create only files  
                ldirNum++;                      // increment leaf dir count
                leafDirs[ldirNum] = ddesc;      // add this to the leaf dir list
                dircnt = 0;                     // no subdirs in a leaf directory
                filcnt = Math.Max(filxc, filesPerDir);  // int needs this many files.
            }
            /////////////////////////////////////////////////////////////////////
            // Allocate the file  or dir tables
            ddesc.fdSDirCnt = dircnt;
            ddesc.fdFileCnt = filcnt;
            /////////////////////////////////////////////////////////////////////
            // get descriptors for existing directories
            if (dircnt > 0)
            {
                fdirs = new fileDesc[dircnt];
                fndcnt = enumDirectory(ddesc, true, fdirs);
                Debug.Assert(fndcnt == dirxc);  // should be the number counted above.
            }
            else
                fdirs = null;
            /////////////////////////////////////////////////////////////////////
            // get descriptors for existing files
            if (filcnt > 0)
            {
                fkids = new fileDesc[filcnt];
                fndcnt = enumDirectory(ddesc, false, fkids);
                Debug.Assert(fndcnt == filxc);   // should be the number counted above.
                /* Create the 'free' list from any unused file entries */
                fileDesc free = null;
                while (filxc < filcnt)
                {
                    fileDesc fdesc = new fileDesc();

                    fdesc.fdParent = ddesc;
                    fdesc.fdIndex = 0;
                    fdesc.fdFileNext = free; free = fdesc;

                    fkids[filxc++] = fdesc;
                }
                ddesc.fdFileFree = free;
            }
            else
                fkids = null;
            ddesc.fdSDirTab = fdirs;
            ddesc.fdFileTab = fkids;
            /////////////////////////////////////////////////////////////////////
            // Create needed sub-directories
            if (curLevel < maxLevel)
            {
                for (int subdir = 1; subdir <= dircnt; subdir++)
                {
                    // Recursively create any needed sub-directories
                    fileCnt = findDirsAndFiles(fileCnt,
                                               subdir,
                                               ddesc,
                                           ref fdirs[subdir - 1],
                                               leafDirs,
                                           ref ldirNum,
                                               curLevel + 1,
                                               maxLevel);
                    // stop when we have enough files.
                    if (fileCnt <= 0)
                        break;
                }   // end sub-directory create loop
            }       // end loop over elemets of this directory.
            //////////////////////////////////////////////////////////////
            // return number of files remaining to be created.
            return fileCnt - filcnt;
        }
        #endregion findDirsAndFiles
        #region enumDirectory
        ///////////////////////////////////////////////////////////////////////////////
        ///////////////////////////////////////////////////////////////////////////////
        /// <summary>
        /// Given a directory name, return a count of the files or sub-directories in
        /// it ('doDirs' says which).
        /// The 'fileDesc[] table' array will be populated with an entry for each such file
        /// or directory if it is not null.
        /// </summary>
        /// <param name="dirDsc">Descriptor of the directory to enumerate</param>
        /// <param name="doDirs">Enumerate sub-dirs if true, otherwise enumerate files</param>
        /// <param name="table">If non-null, entries will be populated with file/sub-dir descriptors</param>
        /// <returns>Number of files/sub-dirs found</returns>
        static int enumDirectory(fileDesc dirDsc,
                                 bool doDirs,
                                 fileDesc[] table)
        {
            string path = dirDsc.fdPath;   // path name to the directory
            string[] directories = null;     // array of directory names in the directory
            FileInfo[] files = null;           // array of file descriptors in the directory
            int count;                  // the number of files or directories in the directory

            ////////////////////////////////////////////////////////////////////////////
            // populate the directory or file array with items in this super-directory
            if (doDirs)
            {   // enumerate the directories
                directories = Directory.GetDirectories(path, "*");
                count = directories.Length;
            }
            else
            {   // enumerate the files.
                files = (new DirectoryInfo(path)).GetFiles();
                count = files.Length;
            }
            //////////////////////////////////////////////////////
            // if  file descriptor table provided fill it in with an entry for each object
            if (table != null)
            {
                //////////////////////////////////////////////////////
                // fill in the file descriptor table with entries for the directories or files.
                for (int i = 0; i < count; i++)
                {
                    string name;
                    fileDesc desc;
                    int fdIndex = 0;
                    /////////////////////////////////////////////////////////
                    // Get the name (it should be of the form Dxxx for a directory, or Fxxx for a file
                    if (doDirs)
                    {
                        name = Path.GetFileName(directories[i]);
                        if (name[0] != 'D')
                        {
                            Console.WriteLine("Bad directory name '" + name + "' encountered, aborting.");
                            System.Environment.Exit(-1);
                        }
                    }
                    else
                    {
                        name = files[i].Name;
                        if (name[0] != 'F')
                        {
                            Console.WriteLine("Bad file name '" + name + "' encountered, aborting.");
                            System.Environment.Exit(-1);
                        }
                    }
                    ///////////////////////////////////////////////////////
                    // Now make a descriptor for it.
                    try
                    {   // insist it is an int following F or D.
                        fdIndex = Convert.ToInt32(name.Substring(1));
                    }
                    catch
                    {
                        Console.WriteLine("Bad file/dir name '" + name + "' encountered, aborting.");
                        System.Environment.Exit(-1);
                    }
                    ///////////////////////////////////////////////////////
                    /// allocate the file descriptor and initialize it.
                    table[i] = desc = new fileDesc();
                    desc.init(doDirs, fdIndex, path + @"\" + name, dirDsc);
                }
            }
            return count;
        }
        #endregion enumDirectory
        #region diskFullness
        /// <summary>
        /// Given a drive letter, computes the fraction (a number between 0 and 1) of
        /// the drive that is full: 0 means drive is empty, 1 means completely full.
        /// </summary>
        /// <param name="drive">Drive letter (including ':')</param>
        /// <returns>Fraction of disk utilized</returns>
        static double diskFullness(string drive)
        {
            DriveInfo dinf = new DriveInfo(drive);

            return 1 - (double)(dinf.AvailableFreeSpace / 1024 / 1024) /
                                (dinf.TotalSize / 1024 / 1024);
        }
        #endregion diskFullness
    }
}
