using System;
using System.IO;
using System.Text;
using System.Diagnostics;
using Microsoft.Win32.SafeHandles;
using System.Runtime.InteropServices;

//////////////////////////////////////////////////////////////////////////////
//
//  Retrieves the number of fragments for a given file.
//
//////////////////////////////////////////////////////////////////////////////

public unsafe class GetFileFrags
{
    #region DLLimports
    [DllImport("KERNEL32", SetLastError = false)]
    static extern int       GetLastError();

    //////////////////////////////////////////////////////////////////////////
    //  Constants and API's for memory allocation:
    //////////////////////////////////////////////////////////////////////////
    const int MEM_RESERVE    = 0x2000;
    const int MEM_COMMIT     = 0x1000;
    const int MEM_RELEASE    = 0x8000;

    const int PAGE_READWRITE = 0x0004;

    [DllImport("KERNEL32", SetLastError = true)]
    static extern void *    VirtualAlloc(void * baseAddr,
                                         uint   size,
                                         uint   allocationType,
                                         uint   protect);

    [DllImport("KERNEL32", SetLastError = true)]
    static extern bool      VirtualFree(void *addr, uint size, int type);
 
    //////////////////////////////////////////////////////////////////////////
    //  Constants and API's to obtain file fragmentation info:
    //////////////////////////////////////////////////////////////////////////

    const int FSCTL_GET_RETRIEVAL_POINTERS = 0x00090073;

    struct  VirtualCluster                  // cluster directory entry
    {
        public  long            NextVcn;    // next cluster
        public  long            Lcn;        // this cluster number
    };
     
    ///  fragments directory;
    struct FRAGS                                // fragments directory;   
    {
        public  uint            ExtentCount;    // count of extents
        public  long            StartingVcn;    // starting cluster numver. 
    };

    [DllImport("KERNEL32", SetLastError = true)]
    static extern bool      DeviceIoControl(SafeFileHandle handle,
                                            int            code,
                                            void *         inpAddr,
                                            uint           inpSize,
                                            void *         outAddr,
                                            uint           outSize,
                                            int  *         retSize,
                                            void *         something);
    #endregion DLLimports
    /// <summary>
    /// Given an open file handle, return the number of fragments allocated for it.
    /// </summary>
    /// <param name="handle">File handle.</param>
    /// <returns>Count of fragments.  If invalid handle, then reports zero fragments.</returns>
    public static uint FileFragments(SafeFileHandle handle)
    {
        FRAGS   *       fragDsc = null;
        long            startVC = 0;
        uint            extSize = 1000;     //Assume most files have fewer than 1000 fragments 

        /////////////////////////////////////////////////////////////////////////////
        /// We don't know how many fragment descriptors we'll need,
        /// so keep trying (with a higher count each time) until successful.
        while(true)
        {
            int  retSize;
            uint buffSiz = (uint)(sizeof(FRAGS) + extSize * sizeof(VirtualCluster));
            ///////////////////////////////////////////////////////////////////////////
            /// Allocate the buffer to hold the fragment descriptors  
            fragDsc = (FRAGS *)VirtualAlloc(null, buffSiz, MEM_RESERVE|MEM_COMMIT, PAGE_READWRITE);
            if  (fragDsc == null)
            {
                Console.WriteLine("Could not allocate file fragment buffer, winerr=" + GetLastError());
                return  0;
            }
            ///////////////////////////////////////////////////////////////////////////
            /// Call the Windows API to retrieve the cluster descriptor table  
            if  (DeviceIoControl(handle,
                                 FSCTL_GET_RETRIEVAL_POINTERS,
                                &startVC, sizeof(long),
                                 fragDsc, buffSiz,
                                &retSize, null))
            {  
                    break;// Success!
            }

            ///////////////////////////////////////////////////////////////////////////
            ///If failed for reasons other than buffer too small, report zero fragments  
            if  (GetLastError() != 234)
            {
                Console.WriteLine("DeviceIoControl() failed, winerr=" + GetLastError());
                return  0;
            }
            ///////////////////////////////////////////////////////////////////////////
            /// Current buffer is too small, free it and allocate double the size  
            VirtualFree(fragDsc, 0, MEM_RELEASE);
            extSize *= 2;
        }

        ///////////////////////////////////////////////////////////////////////////////
        /// Get the total fragment count from the descriptor, free it, and return count 
        uint fragCnt = fragDsc->ExtentCount;
        VirtualFree(fragDsc, 0, MEM_RELEASE);
        return  fragCnt;
    }
}
