﻿using System;
using System.IO;
namespace IOexamples
{
    /// <summary>
    ///  ------------------------------------------------------
    /// <h1>
    /// MainProgram class     File: IOexamples.cs
    /// </h1>
    /// Description: Code demonstrates different styles of and performance of
    /// sequential file IO on Windows NTFS file systems using .NET IO classes.
    /// These are the program examples for the MSR Technical Report by Kukol and Gray
    ///    titled: Sequential File Programming Patterns and Performance with .NET
    ///    The code is downloadable from:
    ///    at URL http://research.microsoft.com/research/downloads/
    ///
    /// This main program invokes and times the following cases:
    ///     Read | Write
    ///     byte-at-a-time | line-at-a-time | 64KB block-at-a-time.
    ///
    /// The file-write-speed test builds the sort benchmark file
    /// specified at http://research.microsoft.com/barc/SortBenchmark/
    ///
    ///  Usage is: IOexamples [fileName [ recordCount] ]
    ///     default file name is:    C:\IO_Examples_temp.txt 
    ///     default record count is                1,000,000
    /// 
    ///  Version: 1.0,  November 2004
    ///
    ///  Authors: Peter Kukol, Jim Gray, Microsoft Corporation
    ///           PeterK@microsoft.com, Gray@Microsoft.com
    ///
    ///  (c) Microsoft Corporation. All rights reserved.
    /// 
    ///</summary>
    class IOexamples
    {
        static void Main(string[] arguments)
        {   //-----------------------------------------------------------
            // Set parameters (or default) and then put out a greeting and summary description.
            //---------------------------------------------------------
            // file name and type default to C:\IO_Examples_temp.txt and 1M records.
            String fileName = @"C:\IO_Examples_temp.txt";
            const long recordSize = 100;  // number of bytes per record
            long records = 1000000;       // default to 1 million records
            // see if command line overrides default.
            try
            {
                if (arguments.Length >= 1)              // if file name provided
                {
                    fileName = arguments[0];            // get file name
                    FileStream f = System.IO.File.Create(fileName); // test that the name is valid
                    fileName = f.Name;                  // get the full path name
                    f.Close();                          // close the file.
                }
                if (arguments.Length >= 2) records  = Convert.ToInt64(arguments[1]);
            }
            catch { // display usage if params are not numbers or a file name
                Console.WriteLine(@"Usage is: IOexamples [fileName [ recordCount] ] ");
                Console.WriteLine(@"          default filename is  C:\IO_Examples_temp.txt ");
                Console.WriteLine(@"          default recordCount is 1000000 ");
                Console.WriteLine(@"          Sequentially reads and writes that many 100-byte records to/from a file stream.");
                Console.WriteLine(@"          using byte-at-a-time, line-at-a-time, and 64KB block-at-a-time.");
                Console.WriteLine(@"          Reports bandwith, cpu time, and elapsed time for each case.");
                Environment.Exit(-1);
            }
            long fileMB = recordSize * records / (1 << 20);  //compute file size.
            Console.WriteLine("{0} version {1} experimenting with a {2:F0} MB file named {3}."
                             + " (c) Microsoft Corporation. All rights reserved.\n", 
                                System.Reflection.Assembly.GetExecutingAssembly().GetName().Name,
                                System.Reflection.Assembly.GetExecutingAssembly().GetName().Version.ToString(),
                                fileMB, fileName);
            //-----------------------------------------------------------
            // timer declares
            double elapsedTime, CPUtime, CPUutilization;
            Timer time = new Timer();

            //----------------------------------------------------------------------------
            // file generation: tests 
            String[] accessMode = { "Byte-At-A-Time", "Line-At-A-Time", "Block-At-A-Time" };
            WriteSortBenchmarkFile w = new WriteSortBenchmarkFile();
            foreach (String mode in accessMode)
            {
                time.timerReset();
                w.writeSortBenchmarkFile(fileName, records, mode);
                elapsedTime = time.timerRead(out CPUtime, out CPUutilization);
                Console.WriteLine("{0} write SortBenchmark {1:F0} MB file in: {2:F2} seconds @ {3:F2} MBps, \n\tCPU: {4:F2},\tUser: {5:F2},\tKernel {6:F2}."
                                 , mode, fileMB, elapsedTime,  fileMB / elapsedTime, CPUtime, time.UserTime, time.KernelTime);
            }
            //----------------------------------------------------------------------------
            // file reads: test the file read speed using the benchmark file
            CountLines c = new CountLines();
            foreach (String mode in accessMode)
            {
                time.timerReset();
                c.countLines(fileName, mode);
                elapsedTime = time.timerRead(out CPUtime, out CPUutilization);
                Console.WriteLine("{0} CountLines of a {1:F0} MB file in: {2:F2} seconds @ {3:F2} MBps, \n\tCPU: {4:F2},\tUser: {5:F2},\tKernel {6:F2}."
                                 , mode, fileMB, elapsedTime, fileMB / elapsedTime, CPUtime, time.UserTime, time.KernelTime);
            }
            //-----------------------------------------------------------------------------
            // cleanup at the end
           File.Delete(fileName);
        }
    }
}
