//////////////////////////////////////////////////////////////////////////////
/// <summary>
/// File:            IOspeed.cs
///
/// Classes:         IOspeed
///
/// Description:     File I/O benchmark program used to measuer the performance of
///                  sequential file IO on Windows NTFS file systems using .NET IO classes.
///                  documented in MSR Technical Report titled: 
///                  "Sequential File Programming Patterns and Performance with .NET"
///                  The code is downloadable from:
///                  at URL http://research.microsoft.com/research/downloads/
///
/// Version:         1.0,  December 2004
///
/// Authors:         Peter Kukol, Jim Gray, Microsoft Corporation
///                  PeteKu@microsoft.com, Gray@Microsoft.com
///
/// (c) Microsoft Corporation 2004. All rights reserved.
/// 
/// IOspeed.exe 
/// 
/// Usage: IOspeed [options] filePath
/// options:
/// -r[fileSize]	read optionally with file size to be created if needed (default=1G)
/// -w[fileSize]	write (default is read) optionally with file size to be written (default=1G)
/// 
/// -t<seconds>	test duration (default=30 seconds)
/// -b<size>	I/O block size (default=64K)
/// -a[count]	use asynch (overlapped) I/O, (Default is sync, default async depth is 4).
/// -d               	disable .NET and NTFS disk caching (memory buffering)
/// -s[seekDistancePercentage] random I/O (default is sequential) 
/// 			optionally average seek distance percentage of filesize, default=100.
/// 
/// -x< fileSize>	create/extend/fill a file to given size
/// -p< fileSize>	like x but preallocate file before writing (reduces fragmentation).
/// 
/// -c              		 touch every byte in the source file (more mememory and cpu load)
/// -q		quiet mode (as opposed to verbose), write data as comma separated list. 
/// 
/// sizes suffixed with K, M, or G are are interpreted as Kilo, Mega, or Giga bytes.
/// 
/// Examples:
///      Iospeed t30 b64 r1G s0 a.dat   	// the default settings, same as next line
///      Iospeed a.dat 	// sequential read a.dat, 64KB requests, buffering for 30 seconds
///      Iospeed t60 -p100M a.dat 	// preallocates 100MB file a.dat, then reads it for 60 seconds
///      Iospeed -t30 -w100M -p a.dat 	// preallocates 100MB file a.dat, then writes it for 30 seconds
///      Iospeed a2 -b256K a.dat 	// 2-deep async read of a.dat in 256KB requests
/// 
/// IOspeed tests read and write bandwidth to a file (or files spread across multiple disks).
/// 
/// The r and w options measure read and write speed respectively. They can be modified 
/// by the t<seconds> option to request a test duration different from the 30 second default, 
/// or with a a<count> option to request asynchronous IO with the specified count of outstanding 
/// IOs, or with the d option to disable .NET and NTFS file caching, or with the b option to 
/// specify an IO request size other than the 64KB default, and with a s<percentage> option to 
/// request random seeks between IOs (-s0) is the sequential default and s100 causes the seek 
/// to go to a random place on the disk. The percentage is a distance control, not a fraction 
/// of sequential-random.  The c option asks the cpu to touch every byte rather than just 
/// discarding the bytes from main memory, making the cpu and memory bandwidth load more realistic.  
/// 
/// The default file size is 1GB.  The fileSize option can override this value. If the file 
/// does not exist it is created with this size. If the file already exists, and it is smaller 
/// than the requested size, it is extended to have this size.  These file extension and the 
/// write tests overwrite the file.
/// 
/// The x and p options test the speed of file extension with and without pre-allocation.
/// If there is a preexisting file with that name, it is deleted and a new empty file is created.
/// x incrementally grows the file by appending to the end (using synchronous writes of the 
/// given block size).  The p option first sets the file length, thereby allocating fewer 
/// fragments and then acts just like the x option. 
/// </summary>
//////////////////////////////////////////////////////////////////////////////

using System;
using System.IO;
using System.Text;
using System.Diagnostics;
using System.Runtime.InteropServices;

using Microsoft.Win32.SafeHandles;

using IOexamples;

//////////////////////////////////////////////////////////////////////////////
//
//  This program can be used to measure the speed of file and network I/O. It
//  has options for reading and writing files (both synchronously and using
//  overlapped / asynchronous I/O), sending data over the network, and many
//  other features.
//
//////////////////////////////////////////////////////////////////////////////

public class IOspeed
{
    #region Globals
    const   long    KILO = 1024;
    const   long    MEGA = 1024 * KILO;
    const   long    GIGA = 1024 * MEGA;

    //////////////////////////////////////////////////////////////////////////
    // Default values for command-line options:
    //////////////////////////////////////////////////////////////////////////

    const   int     DFLT_BLOCK_IN_KB  = 64;             // -b default is 64 KB
    const   long    DFLT_WRTSZ_IN_GB  = 1;              // -x default is 1 GB
    const   int     DFLT_MAXTM_IN_SEC = 30;             // -t default is 30 seconds
    const   int     DFLT_ASYNCH_DEPTH = 4;              // -a default is 4 I/O's

    //////////////////////////////////////////////////////////////////////////
    // Command-line option values:
    //////////////////////////////////////////////////////////////////////////

    static  int     blockSize  = (int)KILO*DFLT_BLOCK_IN_KB; // size of I/O block in bytes

    static  bool    asyncIO   = false;                  // overlapped I/O enabled
    static  int     IOdepth    = DFLT_ASYNCH_DEPTH;     // overlapped I/O count

    static FileAccess mode = FileAccess.Read;           // do writes instead of reads
    static  long    fileSize  = GIGA*DFLT_WRTSZ_IN_GB;  // Desired size of file (default 1GB)

    static  int     randPct    = 0;                     // randomness percentage

    static  bool    noBuff     = false;                 // disable buffering
    static  bool    doCopy     = false;                 // copy every byte (increase memory load)

    static  bool    quietMode  = false;                 // 'quiet' mode for spreadsheet gen

    static bool     fileXtest = false;                  // do file extend tests
    static bool     preAllocate = false;                // when writing or extending a file, 
    // first preallocate it so that it has fewer fragments. 
 
    //////////////////////////////////////////////////////////////////////////

    static  uint    sectorSize = 512;                   // target volume sector size

     
    /////////////////////////////////////////////////////////////////////////
    /// Timing code
    static double   max_time = 1000 * DFLT_MAXTM_IN_SEC;// don't run longer than this
    static long     time_beg = 0;
    static Timer    timeIt   = new Timer();

    #endregion Globals
    #region DLLimports
    [DllImport("KERNEL32", SetLastError = true)]
    static extern bool FlushFileBuffers(SafeFileHandle handle);
    #endregion DLLimports
    #region commandHandling
    /// <summary>
    /// Display usage string. 
    /// </summary>
    static void usage()
    {
        Console.WriteLine("{0} version {1} measures file IO performance based on specified parameters."
                         + " (c) Microsoft Corporation. All rights reserved.\n",
                            System.Reflection.Assembly.GetExecutingAssembly().GetName().Name,
                            System.Reflection.Assembly.GetExecutingAssembly().GetName().Version.ToString());
        Console.WriteLine(""); 
        Console.WriteLine("Usage: IOspeed [options] file_path");
        Console.WriteLine("");
        Console.WriteLine("Available options:");
        Console.WriteLine("");
        Console.WriteLine("    -r[fileSize]     read test, optionally with file size to be created if needed (default=" + DFLT_WRTSZ_IN_GB + "G)");
        Console.WriteLine("    -w[fileSize]     write test (default is read), optionally with file size to be written (default=" + DFLT_WRTSZ_IN_GB + "G)");
        Console.WriteLine("");
        Console.WriteLine("    -t<seconds>      test duration (default=" + DFLT_MAXTM_IN_SEC + " seconds)");
        Console.WriteLine("    -b<size>         I/O block size (default=" + DFLT_BLOCK_IN_KB + "K)");
        Console.WriteLine("    -a[count]        use asynch (overlapped) I/O, (Default is sync, default async depth is " + DFLT_ASYNCH_DEPTH + ").");
        Console.WriteLine("    -d               disable .NET and NTFS disk caching (memory buffering)");
        Console.WriteLine("");
        Console.WriteLine("    -s[seekDistancePercentage] random I/O instead of sequential, optionally average seek distance percentage of filesize, default=100.");
        Console.WriteLine("    -x<size>         create/extend/fill a file to given size");
        Console.WriteLine("    -p<size>         like x but preallocate file before writing (reduces fragmentation).");
        Console.WriteLine("");
        Console.WriteLine("    -c               touch every byte in the source file (more mememory and cpu load)");
        Console.WriteLine("    -q               quite mode, write data in a format convenient for programs"); 
        Console.WriteLine("");
        Console.WriteLine("sizes suffixed with K, M, or G are are interpreted as Kilo, Mega, or Giga bytes.");
        Console.WriteLine("");
        Console.WriteLine("Examples:");
        Console.WriteLine("     Iospeed t30 b64 r1G s0 a.dat   // the default settings, same as next line");
        Console.WriteLine("     Iospeed a.dat                      // sequential read a.dat, 64KB requests, buffering for 30 seconds");
        Console.WriteLine("     Iospeed t60 -p100M a.dat          // preallocates 100MB file a.dat, then reads it for 60 seconds");
        Console.WriteLine("     Iospeed -t30 -w100M -p a.dat       // preallocates 100MB file a.dat, then writes it for 30 seconds");
        Console.WriteLine("     Iospeed a2 -b256K a.dat           // 2-deep async read of a.dat in 256KB requests");
        Console.WriteLine("");
    }
    /// <summary>
    /// Scale and int by Kilo, Mega, or Giga if the string has such a suffix. 
    /// </summary>
    /// <param name="size">the string has the syntax -x####[K|M|G] </param>
    /// <returns>returns size, or -1 if there is an error.</returns>
    static long parseSize(string size)
    {  
        long iSize;
        
        try
        {
                      if (size[size.Length - 1] == 'K') iSize = Convert.ToInt64(size.TrimEnd('K')) * KILO;
                 else if (size[size.Length - 1] == 'M') iSize = Convert.ToInt64(size.TrimEnd('M')) * MEGA;
                 else if (size[size.Length - 1] == 'G') iSize = Convert.ToInt64(size.TrimEnd('G')) * GIGA;
                 else                                   iSize = Convert.ToInt64(size);
        }
        catch (Exception e)
        {
            Console.WriteLine(e.ToString());
            return -1;
        }
        return iSize;
    }
    /// <summary>
    /// Convert "+" and "-" toggle options to boolean
    /// </summary>
    /// <param name="optStr">the "+" or "-" or null string (== "+") </param>
    /// <param name="optChar">the option being controlled</param>
    /// <returns>true if option "+", false if option "-", exit otherwise.</returns>
    static bool checkToggleOpt(string optStr, char optChar)
    {
        if (optStr.Equals("")) return true;
        if (optStr.Equals("+")) return true;
        if (optStr.Equals("-")) return false;

        Console.WriteLine("Invalid '" + optChar + "' option: use -" + optChar + " to enable, -" + optChar + "- to disable");
        usage();
        System.Environment.Exit(-1);
        return false;
    }
    /// <summary>
    /// parseCommand(string [] argv) parses the command string
    /// </summary>
    /// <returns>true if command string OK, false otherwise exit otherwise.</returns>
    static bool parseCommand(string[] args)
    {
        int argCnt = 0;
        ////////////////////////////////////////////////////////////
        /// make sure there at least one param
        if (args.Length == 0) goto badParams;
        ///////////////////////////////////////////////////////////
        /// process each option, except the last one (file name)
        foreach (string arg in args)
        {
            // last argument does is path name, not a flag so end flag loop
            if (arg[0] != '-' && arg[0] != '/') break;
            /// if not at least 2 long, then it is a syntax error
            if (arg.Length < 2)
            {
                Console.WriteLine("Invalid option: '" + arg + "'");
                goto badParams;
            }
            string argVal = arg.Substring(2); // get flag suffix if it exists. 

            switch (arg[1])
            {
                /////////////////////////////////////////////////
                /// -a[depth] means asyncronous IO with that depth
                case 'a':
                    asyncIO = true;
                    if (argVal == "")
                    {
                        IOdepth = DFLT_ASYNCH_DEPTH;
                        break;
                    }
                    else
                    {
                        IOdepth = (int) parseSize(argVal);
                        if (0 < IOdepth && IOdepth < 1000 ) break;
                        Console.WriteLine("Invalid async I/O depth argument: '" + arg + "'");
                        goto badParams;
                    }
                /////////////////////////////////////////////////
                /// -b[blockSize] the request transfer size. 
                case 'b':
                    blockSize  = (int) parseSize(argVal);
                    if (0 < blockSize) break;
                    Console.WriteLine("Invalid block size specification: '" + arg + "'");
                    goto badParams;
                /////////////////////////////////////////////////
                /// -c requests copying every byte of source file to increase cpu and memory load.  
                case 'c':
                    doCopy = true;
                    break;
                /////////////////////////////////////////////////
                /// -d disables software caching 
                case 'd':
                    noBuff = true;
                    break;  
                /////////////////////////////////////////////////
                /// -q enables 'quiet' mode
                case 'q':
                    quietMode = true;
                    break;
                /////////////////////////////////////////////////
                /// -s[seekDistancePercentage] random I/O instead of sequential, 
                /// with optional average seek distance percentage, default is 100%.
                case 's':
                    randPct = 100;                      // the default value
                    if (argVal == "") break;            // take default if no override
                    randPct = (int) parseSize(argVal);
                    if (0 <= randPct  &&  randPct <= 100) break;
                    Console.WriteLine("The value of " + arg + " must be between 0 and 100.");
                    goto badParams;
                /////////////////////////////////////////////////
                /// -t<seconds>  running time in seconds
                case 't':
                    max_time = 1000 * (int) parseSize(argVal);
                    if (max_time > 0) break;
                    Console.WriteLine("Invalid time argument:  '" + arg + "'");
                    goto badParams;
                /////////////////////////////////////////////////
                /// -r[fileSize] specifies a read-file test and optionally specifies file size.
                case 'r':                       //  this is the default
                    mode = FileAccess.Read;     // do read test, default of 1GB size already set.
                    if (argVal == "") break;
                    fileSize = parseSize(argVal);
                    if (fileSize > 0) break;
                    Console.WriteLine("Invalid '-w' option (expected '-w' or '-w<size>'):  '" + arg + "'");
                    goto badParams;
                /////////////////////////////////////////////////
                /// -w[fileSize] specifies a write-file test and optionally specifies file size.
                case 'w':
                    mode = FileAccess.Write;   // do write test, default of 1GB size already set.
                    if (argVal == "") break;
                    fileSize = parseSize(argVal);
                    if  (fileSize > 0) break;
                    Console.WriteLine("Invalid '-w' option (expected '-w' or '-w<size>'):  '" + arg + "'");
                    goto badParams;
                /////////////////////////////////////////////////
                /// -x<size> runs incremental 'file extend' test
                case 'x':
                    fileXtest = true;
                    if (argVal == "") break;
                    fileSize = parseSize(argVal);
                    if (fileSize > 0) break;
                    Console.WriteLine("Invalid file extend size argument (expected '-x<size>'): '" + arg + "'");
                    goto badParams;
                /////////////////////////////////////////////////
                /// -p<size> preallocate 'file extend' test
                case 'p':
                    preAllocate = true;
                    fileXtest = true;
                    if (argVal == "") break;
                    fileSize = parseSize(argVal);
                    if (fileSize > 0) break;
                    Console.WriteLine("Invalid file extend size argument (expected '-x<size>'): '" + arg + "'");
                    break;
                /////////////////////////////////////////////////
                /// display usage
                case '?': case 'h':
                    // usage(); parent will do this
                    goto badParams;
                /////////////////////////////////////////////////
                /// unknown switch type.
                default:
                    Console.WriteLine("Invalid option: '" + arg + "'");
                    goto badParams;
            }
            /// processed a good argument, increment the count. 
            argCnt++;
        }
        //////////////////////////////////////////////////////////////////////////////////////////
        /// After all parameters are parsed, there should be a  single directory/drive argument  
        if (argCnt != args.Length - 1)
        {
            Console.WriteLine("Expected a single file argument at the end, found " + argCnt + " argument(s) instead.");
            goto badParams;
        }
        //////////////////////////////////////////////////////////////////////////////////////////
        /// check the mutual consistency of the options.  
        if (fileXtest)
        {
            if (asyncIO || doCopy || noBuff || randPct != 0)
            {
                Console.WriteLine("Option combination not supported,  -x and -p do not allow -a, -c, -d or -s ");
                goto badParams;
            }
        }
        //////////////////////////////////////////////////////////////////////////
        /// Everything was OK.  
        return true;

        //////////////////////////////////////////////////////////////////////////
        /// come here when things are bad.
   badParams:
        usage();
        return false;
    }
    #endregion commandHandling
    #region size2str
    //////////////////////////////////////////////////////////////////////////
    /// <summary>
    /// Given a byte count, create a string with a user-friendly representation.
    ///  like 30.24 GB rather than 32469952758
    /// </summary>
    /// <param name="speed">the float byte count.</param>
    /// <returns>a string of the form "###.00 KB "</returns>
     static string size2str(double speed)
    {
        if (speed < 1024) return speed.ToString("###.##") + " bytes";
        speed /= 1024;
        if (speed < 1024) return speed.ToString("###.##") + " KB";
        speed /= 1024;
        if (speed < 1024) return speed.ToString("###.##") + " MB";
        speed /= 1024;
                          return speed.ToString("###.##") + " GB";
                      }
    #endregion size2str
    #region fileExtend
    /// <summary>
    /// Extends (or creates) a file if a given size, returning the number of fragments.
    /// Just counts fragments if file is already big enough.
    /// </summary>
    /// <param name="fileName">name of the file to extend</param>
    /// <param name="fileSize">the desired file size (in bytes).</param>
    /// <returns>returns a count fo the number for fragments, or -1 if there is a problem. </returns>
    static int fileExtend(string fileName, long fileSize )
    {
        try
        {   
            using (FileStream fs = new FileStream(fileName, FileMode.OpenOrCreate,
                                                    FileAccess.Write, FileShare.None) )
            {
                /////////////////////////////////////////////////////////////////////////
                //// if file not big enough, grow it, otherwise just count fragments.
                if (fs.Length < fileSize)
                {  
                    /////////////////////////////////////////////////////////////////////////
                    //// pre-allocate the full size if that option requested.
                    if (preAllocate)
                    { 
                        fs.SetLength(fileSize);
                    }
                    ////////////////////////////////////////////////////////////////////////
                    /// write the file to the desired length. 
                    byte[] incData = new byte[blockSize];       // allocate the write buffer
                    fs.Seek(0, SeekOrigin.Begin);               // position at start of file
                    int sizeAdd;                                // size of next write
                    for (long fileOffset = 0; fileOffset < fileSize; fileOffset += sizeAdd)
                    {
                        sizeAdd = (int)Math.Min(fileSize - fileOffset, 
                                                blockSize); // size of next write
                        fs.Write(incData, 0, sizeAdd);          // do the write 
                    }
                    ////////////////////////////////////////////////////////////////////////
                    /// Write complete, flush the buffers and get the fragmentation count. 
                    FlushFileBuffers(fs.SafeFileHandle);
                }
                ////////////////////////////////////////////////////////////////////////
                /// Return fragmentation count.
                return (int) GetFileFrags.FileFragments(fs.SafeFileHandle);
            }
        }
        ////////////////////////////////////////////////////////////////////////
        /// if the file access or storage allocation fails, control will come here. 
        catch (Exception e)
        {
            Console.WriteLine("ERROR while creating file: \n" + e.ToString());
            return -1;
        }
    }
    #endregion FileExtend
    #region SynchronousAccess
    /// <summary>
    /// Synchronously read or write a given file in the designated way (random / sequential) 
    /// for the designated time.
    /// Returns the  number of bytes and I/O's.
    /// Starts the timer time_it when the experiment starts. 
    /// </summary>
    /// <param name="fileName">name of the file to read/write</param>
    /// <param name="mode"> FileAccess.Read or FileAccess.Write</param>
    /// <param name="doneBytes">How many bytes were read/write</param>
    /// <param name="doneIOs">How many IOs were issued.</param>
    /// <returns>boolean: true: it worked, false: outputs are bogus. </returns>
    static Boolean SynchronousAccess(string fileName, out long doneBytes, out long doneIOs)
    {
        ////////////////////////////////////////////////////////////////////////
        byte[]          IObuff   = null;    // buffer for IOs
        FileStream      fs       = null;    // all IO will be done against this stream
        OpenFileNoBuff  of       = null;    // if IO is unbuffered, this is the unbuffred NTFS handle.
        long            fileOffset = 0;     // offset into file
        int             byteCount = 0;      // count of bytes returned by a read or returned by write. 
        double          maxRandOffset;      // Random IOs will seek a random distance witin this max
        Random          randGen;            // random number source. 
        int             checkSum = 0;       // to fake touching the bytes, we checksum them

        ///////////////////////////////////////////////////////////////////////
        /// Intialize 'out' parameters, allocate write buffer
        doneBytes = 0; doneIOs = 0; 
        try
        {
            if  (blockSize > 1) IObuff = new byte[blockSize];
        }
        catch(OutOfMemoryException)
        {
            Console.WriteLine("ERROR: Could not allocate " + blockSize + " bytes for the I/O buffer!");
            return  false;
        }

        ///////////////////////////////////////////////////////////////////////
        /// Open the file and fileStream  with the correct buffering option.
        try
        {   ///////////////////////////////////////////////////////////////////
            ///  The unbuffered case.
            if  (noBuff)
            { 
                of = new OpenFileNoBuff();
                fs = of.openStream (fileName,
                                    FileMode.OpenOrCreate,
                                    FileAccess.ReadWrite,
                                    FileShare.None,
                                    (randPct == 0),
                                    false,
                                    blockSize);
                if  (fs == null)
                {   /// Give diagnostic and return false if can't open file. 
                    Console.WriteLine("ERROR: Could not open file '" + fileName + "', please check the path and try again.");
                    return  false;
                }
            }
            else
            {   ///////////////////////////////////////////////////////////////////
                ///  The 'normal' (i.e. buffered) stream  
                fs = new FileStream(fileName,
                                    FileMode.OpenOrCreate,
                                    FileAccess.ReadWrite,
                                    FileShare.None,
                                    blockSize,
                                    (randPct == 0) ? FileOptions.SequentialScan
                                                   : FileOptions.RandomAccess);
            }
            ///////////////////////////////////////////////////////////////////////
            ///  Get file demographics, and extend it if necessary. 
            fileSize = Math.Max(fs.Length,fileSize);
            /// Compute the max. distance (wrap around) to move for random I/O */
            maxRandOffset = fileSize * randPct / 100.0;
            randGen  = new Random(127);
            /////////////////////////////////////////////////////////////////////////
            /// Environoment set up, now measure the bandwidth. 
            time_beg = Environment.TickCount;       // start the clocks
            timeIt.timerReset();
            long delta = Environment.TickCount - time_beg;
            while (Environment.TickCount - time_beg <= max_time)
            {
                delta = Environment.TickCount - time_beg;
                /// If random IO, position at a random offset in the file
                /// relative to current position, and biased by maxRandOffset
                if  (randPct != 0)
                {
                    fileOffset = (long)(fs.Position + maxRandOffset * randGen.NextDouble()) % fileSize;
                    fileOffset &= ~(long)(blockSize-1);  // align to blocksize boundary
                    fs.Seek(fileOffset, SeekOrigin.Begin);
                }
                /// Actual work
                if  (blockSize == 1)
                {   /// Byte-at-a-time.
                    if (mode == FileAccess.Read)
                    {
                        int nextByte = fs.ReadByte();   // Read the next byte and check for end of file  
                        byteCount = (nextByte >= 0) ? 1 : 0; // count of bytes read, negative means end of file
                        if (doCopy && byteCount == 1) checkSum += nextByte; // phoney byte checksum, to show data read.
                    }
                    else
                    {
                        fs.WriteByte((byte)fileOffset);   // write the next byte with low byte of offset
                        byteCount = 1;
                    }
                }
                else
                {   /// Block-at-a-time.
                    if (mode == FileAccess.Read)
                    {/// Block reader. 
                        byteCount = fs.Read(IObuff, 0, blockSize); // Read the next buffer  
                        if (doCopy)
                        {
                            for (int i = 0; i < byteCount; i++)
                                checkSum += IObuff[i];  // phoney byte checksum, just to show data read. 
                        }
                    }
                    else {
                        byteCount = (int)Math.Min(blockSize, fileSize - fileOffset);
                        if (doCopy)
                        {
                            for (int i = 0; i < byteCount; i++)
                                 IObuff[i] = (byte) i;  // phoney byte checksum, just to show data read. 
                        }
                        fs.Write(IObuff, 0, byteCount); // write the next buffer  
                    }
                }
                /////////////////////////////////////////////////////////////////
                /// Request complete, count I/O's and bytes we've processed  
                doneIOs   += 1;
                doneBytes += byteCount;
                fileOffset = fileOffset + byteCount;
                if (fileOffset >= fileSize)
                    { 
                    fileOffset = 0;
                    fs.Seek(fileOffset, SeekOrigin.Begin); 
                    }
            }
            /////////////////////////////////////////////////////////////////////
            /// end of a scan of the file,  if writing, flush the buffers
            if (mode == FileAccess.Write && !noBuff)
                FlushFileBuffers(fs.SafeFileHandle);
            /////////////////////////////////////////////////////////////////////
            /// return 
            return true;
        }
        catch(Exception e)
        {
            /// If something goes wrong, display the error and return */
            Console.WriteLine(e.ToString());
        }
        finally
        {   ////////////////////////////////////////////////////////////////////
            /// Make sure we close the stream no matter how we return  
            if  (of != null) of.closeStream();
            if  (fs != null) fs.Close();
        }
        return false;
    }
    #endregion SynchronousAccess
    #region AsyncFileIO
    //////////////////////////////////////////////////////////////////////////
    //
    //  Asynchronous file I/O logic starts here:
    //
    //////////////////////////////////////////////////////////////////////////
   /// <summary>
   /// Async File IO descriptor  
   /// </summary>
   class fileIOdesc
    {
        public int          IOindex;            // ?????????????????????????
        public bool         IOdone;             // flag says we processed the IO
        public FileAccess   IOmode;             // mode is read / write
        public int          IOsize;             // bytes reqested
        public int          IObufferOffset;     // offset into the IO buffer
        public Stream       IOstream;           // stream that got the request
        public byte    []   IObuffer;           // buffer that holds the data
    }
    /// <summary>
    /// Gloabal variables for async code. 
    /// </summary>
    static  Object  fileIOdoneLock;         // lock to synchronoize updates to shared data. 
    static  uint    fileIOdoneCnt;          // count of completed IOs
    #region fileIOdone
    //////////////////////////////////////////////////////////////////////////
    /// <summary>
    /// IO completion delegate called by .NET whenever an asynchronous file I/O completes.
    ///         execute the EndRead()/EndWrte() method, mark the IO as complete, increment the IO count.
    /// </summary>
    /// <param name="ar">fileIO descriptor associated with the completed IO</param>
    static void fileIOdone(IAsyncResult ar)
    {
        fileIOdesc      IOdesc = (fileIOdesc)ar.AsyncState;
        Stream          stream = IOdesc.IOstream;
        //////////////////////////////////////////////////////////////////////
        /// perform the EndRead()/EndWrite() on request
        if (IOdesc.IOmode == FileAccess.Write)
        {
            stream.EndWrite(ar);
        }
        else
        {
            int size = stream.EndRead(ar);
            if (size != IOdesc.IOsize) /* Make sure we processed the expected number of bytes */
                throw new Exception("ERROR: Completed file I/O processed " + size + " bytes instead of " + IOdesc.IOsize);
        }
        ////////////////////////////////////////////////////////////////////
        /// Signal mainline that this I/O request as complete and increment counters. 
        Debug.Assert(IOdesc.IOdone == false);
        IOdesc.IOdone = true;
        lock(fileIOdoneLock)
        {
            fileIOdoneCnt++;
            System.Threading.Monitor.Pulse(fileIOdoneLock);
        }
    }
    #endregion fileIOdone
    #region AsynchronousAccess
    /// <summary>
    /// Asynchronously read or write a given file in the designated way (random / sequential) 
    /// for the designated time.
    /// Returns the  number of bytes and I/O's 
    /// Starts the timer time_it when the experiment starts. 
    /// </summary>
    /// <param name="fileName">name of the file to read/write</param>
    /// <param name="mode"> FileAccess.Read or FileAccess.Write</param>
    /// <param name="doneBytes">How many bytes were read/write</param>
    /// <param name="doneIOs">How many IOs were issued.</param>
    /// <returns>boolean: true: it worked, false: outputs are bogus. </returns>  
    static Boolean AsynchronousAccess(string fileName,
                           out long doneBytes, out long doneIOs)
    {
        bool starting   = true;     // we are starting to issue IOs.
        uint pending    = 0;        // this many are pending 
        long fileOffset = 0;        // this is where we are in the file. 
        double maxRandOffset;       // Random IOs will seek a random distance witin this max
        Random randGen;             // random number source. 

        FileStream      fs = null;  // the file stream
        OpenFileNoBuff  of = null;  // the unbuffered NTFS file handle.  
        fileIOdesc[]    IOtable;    // describes each of async# IOs.
        int             IOnum = 0;  // index into the IOtable
        int             nextIO = 0; // during the starting phase, tracks next free buffer.
        byte  []        IObuff;     // holds  async# of buffers of blockSize
        AsyncCallback   IOcallback = new AsyncCallback(fileIOdone);  // handles EndRead(), EndWrite()
        byte toss = 0;              // target of byte-touching code for -c option
        ////////////////////////////////////////////////////////////////////////////
        /// All 'out' parameters must be initialized */
        doneBytes = 0; doneIOs = 0; 
        ////////////////////////////////////////////////////////////////////////////
        /// Allocate all the tables we need 
        try
        {
            IOtable = new fileIOdesc[IOdepth];          // IO descriptors
            IObuff = new byte[blockSize * IOdepth];    // IO buffers
        }
        catch (OutOfMemoryException)
        {
            Console.WriteLine("ERROR: Could not allocate memory for I/O buffers / tables");
            return false;
        }
        ////////////////////////////////////////////////////////////////////////////
        /// Open the file and create a stream
        try
        {
            if (noBuff)
            { /* We're supposed to use un-buffered I/O */
                of = new OpenFileNoBuff();
                fs = of.openStream(fileName,
                                    FileMode.OpenOrCreate,
                                    FileAccess.ReadWrite,
                                    FileShare.None,
                                    true,
                                    true,
                                    blockSize);
            }
            else
            {
                fs = new FileStream(fileName,
                                    FileMode.OpenOrCreate,
                                    FileAccess.ReadWrite,
                                    FileShare.None,
                                    blockSize,
                                    FileOptions.Asynchronous);
            }
            if (fs == null)
            {
                Console.WriteLine("ERROR: Could not open file '" + fileName + "', please check the path and try again.");
                return false;
            }
            ///////////////////////////////////////////////////////////////////////
            ///  Get file demographics
            fileSize = Math.Max(fs.Length, fileSize);
            ///////////////////////////////////////////////////////////////////////
            /// Compute the max. distance (wrap around) to move for random I/O */
            maxRandOffset = fileSize * randPct / 100.0;
            randGen = new Random(127);
            ///////////////////////////////////////////////////////////////////////
            /// initialize the global call-back variables  
            fileIOdoneLock = new Object();
            fileIOdoneCnt = 0;
            //////////////////////////////////////////////////////////////////////
            /// Fill in the I/O descriptor table  
            for (IOnum = 0; IOnum < IOdepth; IOnum++)
            {
                fileIOdesc IOdsc = new fileIOdesc();
                IOdsc.IOindex = IOnum;                    // index of this request in the array
                IOdsc.IOdone = false;                     // flag indicates we did an EndRead() or EndWrite()
                IOdsc.IOmode = mode;                      // Read or Write
                IOdsc.IOsize = 0;                         // will be filled in when request is made
                IOdsc.IOstream = fs;                      // stream for this request
                IOdsc.IObufferOffset = IOnum * blockSize; // each request gets a part of the buffer
                IOdsc.IObuffer = IObuff;                  // base address of IO buffer
                IOtable[IOnum] = IOdsc;                   // add formatted request to the table. 
            }
            /////////////////////////////////////////////////////////////////////////
            /// Environoment set up, now measure the bandwidth. 
            time_beg = Environment.TickCount;       // start the clocks
            timeIt.timerReset();
            //////////////////////////////////////////////////////////////////////////
            /// loop till the time is up.  Each time round the loop issues a new IO.
            /// until max_time is reached.  When past the start and no IO pending, we can exit
            while (starting || (pending > 0))
            {

                int IOsize;
                /// While starting, issue overlapping requests using the first IOdepth buffers.   
                if (starting)
                {
                    IOnum = nextIO;                 // pick next IO
                    if (++nextIO == IOdepth)        // increment 
                        starting = false;           // starting ends when have launced IOdepth IOs
                }
                else
                {/// Past the starting phase, find a completed IO and re-issue it.
                    lock (fileIOdoneLock)           // test the fileIOdoneCnt
                    {
                        while (fileIOdoneCnt == 0)  // if it is zero, release lock and wait
                            System.Threading.Monitor.Wait(fileIOdoneLock);
                        fileIOdoneCnt--;            // it is non-zero, decrement by one
                    }
                    /// Find a completed I/O request 
                    for (IOnum = 0; IOnum < IOdepth; IOnum++)
                    {
                        if (IOtable[IOnum].IOdone)  // found a completed request    
                            break;
                    }
                    //////////////////////////////////////////////////////////////////
                    /// found a completed request,  record completion and use it for next operation
                    Debug.Assert(IOtable[IOnum].IOdone == true);
                    IOtable[IOnum].IOdone = false;          // mark it complete
                    pending--;                              // reduce pending count
                    doneIOs += 1;                           // increment IO count
                    doneBytes += IOtable[IOnum].IOsize;     // add processed bytes.
                    if (mode == FileAccess.Write && doCopy) // touch the bytes to generate cpu/memory load. 
                        for (int i = 0; i < IOtable[IOnum].IOsize; i++)
                            toss = IObuff[IOtable[IOnum].IObufferOffset + i];
                }
                //////////////////////////////////////////////////////////////////////
                /// when max_time reached, stop issuing new IOs
                if (Environment.TickCount - time_beg >= max_time)
                {
                    starting = false;
                    continue;
                }
                //////////////////////////////////////////////////////////////////////
                /// Compute offset for next IO, random is random spot in file
                if (randPct != 0)
                {
                    fileOffset = (long)(fs.Position + maxRandOffset * randGen.NextDouble()) % fileSize;
                    fileOffset &= ~(long)(blockSize - 1);  // align to blocksize boundary
                    fs.Seek(fileOffset, SeekOrigin.Begin);
                }
                else
                {   /// Sequential is next block in file with wrap-around to start of file
                    fileOffset += blockSize;
                    if (fileOffset >= fileSize)
                    {   /// the wrap-around case when file is filled. 
                        fileOffset = 0;
                        fs.Seek(0, SeekOrigin.Begin);
                    }
                }
                /// Write blocksize (or less if near end of file)
                IOsize = (int)Math.Min(blockSize, fileSize - fileOffset);
                fileIOdesc IOdesc = IOtable[IOnum];
                IOdesc.IOsize = IOsize;
                /// Issue next Async IO request
                if (mode == FileAccess.Write)   // issue a write
                {   // if requested to touch bytes, do them before the write
                    if (doCopy)  // phoney byte checksum, just to show data read
                        for (int i = 0; i < IOsize; i++) IObuff[IOdesc.IObufferOffset + i] = (byte)i;
                    fs.BeginWrite(IObuff,       // base address of   I/O buffer
                        IOdesc.IObufferOffset,  // offset within I/O buffer 
                        IOsize,                 // how many bytes to write
                        IOcallback,             // function to call when I/O completes
                        IOdesc);                // cookie to pass to completion function
                }
                else // issue a read
                {   // touch bytes happens after read completes
                    fs.BeginRead(IObuff,        // base address of   I/O buffer
                        IOdesc.IObufferOffset,  // offset within I/O buffer
                        IOsize,                 // how many bytes to read
                        IOcallback,             // function to call when I/O completes
                        IOdesc);                // cookie to pass to completion function
                }
                /// We have one more I/O request 'in flight' */
                pending++;
            }
        /////////////////////////////////////////////////////////////////////
        /// end work, if writing, flush the buffers
        if (mode == FileAccess.Write && !noBuff)
                FlushFileBuffers(fs.SafeFileHandle);
        /////////////////////////////////////////////////////////////////////
        /// return 
        return true;
        }
        catch (Exception e)
        {
            /// If something goes wrong, display the error and return */
            Console.WriteLine(e.ToString());
        }
        finally
        {   ////////////////////////////////////////////////////////////////////
            /// Make sure we close the stream no matter how we return  
            if (of != null) of.closeStream();
            if (fs != null) fs.Close();
        }
        return false;
    }
    #endregion AsynchronousAccess
    #endregion AsyncFileIO
#region Main
    //////////////////////////////////////////////////////////////////////////
    /// IOspeed main routine.  Parses input, launches extend, read, write tests.
    /// Generates summary statistics. 
    //////////////////////////////////////////////////////////////////////////
    public static  void Main(String[] args)
    {
        string          fileName;               // file we are operating on
        long            doneBytes = 0;          // bytes processed
        long            doneIOs   = 0 ;         // IOs processed.
        int             fragments = 0;          // file fragments
        double elapsedTime, CPUtime, CPUutilization; // timer variables
        //////////////////////////////////////////////////////////////////////
        /// parse the command line, and exit if there are any problems.
        if (!parseCommand(args)) return;
        if (!quietMode)
        {
            Console.WriteLine("{0} version {1}."
                             + " (c) Microsoft Corporation. All rights reserved.\n",
                                System.Reflection.Assembly.GetExecutingAssembly().GetName().Name,
                                System.Reflection.Assembly.GetExecutingAssembly().GetName().Version.ToString());
        }
        //////////////////////////////////////////////////////////////////////
        /// get the file name we are working on.
        fileName = args[args.Length-1];
        //////////////////////////////////////////////////////////////////////
        /// If it is a "file extend" test, then time it and describe the fragments and speeds. 
        if (fileXtest)
        {
            if (File.Exists(fileName)) File.Delete(fileName);   // delete the file if it exists
            timeIt.timerReset();                                // start the clock
            fragments = fileExtend(fileName, fileSize);         // extend the file
            elapsedTime = timeIt.timerRead(out CPUtime, out CPUutilization);    // stop the clock
            if  (quietMode)
                Console.WriteLine("File extend: {0}, {1:F2}, {2:F}, {3:F2}, {4} ",
                            fileSize, elapsedTime, (fileSize / elapsedTime), CPUtime, fragments);
            else
                Console.WriteLine(@"file extend: {0} time = {1:F2} seconds, speed = {2}/s, cpuSeconds = {3:F2}, fragCnt = {4}",
                                    size2str(fileSize), elapsedTime, size2str(fileSize / elapsedTime), CPUtime, fragments);
            return;
        }
        //////////////////////////////////////////////////////////////////////
        /// create the destination file of the requested size 
        preAllocate = true;                         // do preallocate to minimize fragments
        fragments = fileExtend(fileName, fileSize); // do the extension
        if (fragments < 1)                          // if there are no fragments, extension failed.
        {                                           // give diagnostic and exit
            usage(); return;
        }
        //////////////////////////////////////////////////////////////////////
        /// For unbuffered tests, make sure the buffer size is a multiple of the sector size. 
        if (noBuff)
        {
            sectorSize = GetDriveSectorSize.driveSectorSize(fileName);
            if ((blockSize & (sectorSize - 1)) != 0)
            {
                Console.WriteLine("ERROR: With -d, the file block size must be a multiple of sector size {0}.", sectorSize);
                usage(); return;
            }
            //// Un-buffered file I/O must process multiples of sector size 
            fileSize = (fileSize+(sectorSize - 1))&(~(long)(sectorSize - 1)); 
        }
        //////////////////////////////////////////////////////////////////////
        /// run the appropriate test, it will start the timer when it is ready 
        Boolean worked = false;
        if  (asyncIO)
            worked = AsynchronousAccess(fileName, out doneBytes, out doneIOs);
        else
            worked = SynchronousAccess( fileName, out doneBytes, out doneIOs);
        //////////////////////////////////////////////////////////////////////
        /// time the test. 
        elapsedTime = timeIt.timerRead(out CPUtime, out CPUutilization);    // stop the clock
        if (elapsedTime > 0 && worked)
        {
            if (quietMode)
                Console.WriteLine("{0}, {1:F2}, {2:F2}, {3:F}, {4:F1}, {5} ",
                            blockSize, elapsedTime, 100*CPUutilization, (doneBytes / elapsedTime), (doneIOs / elapsedTime), fragments);
            else
                Console.WriteLine(@"blocksize = {0}, time = {1:F2} seconds, cpu = {2:F2}% speed = {3}/s, IO = {4:F0}/s, fragments = {5}.",
                            size2str(blockSize), elapsedTime, 100 * CPUutilization,
                            size2str(doneBytes / elapsedTime), doneIOs / elapsedTime, fragments);
        }
    }
    #endregion Main
}

