using System;
using System.IO;
using System.Runtime.InteropServices;
using Microsoft.Win32.SafeHandles;

//////////////////////////////////////////////////////////////////////////////
//
//  Open a fileStream with NTFS buffering disabled.
//
//////////////////////////////////////////////////////////////////////////////
public class OpenFileNoBuff
{
    #region DLLimport
    const int FILE_FLAG_NO_BUFFERING    = unchecked((int)0x20000000);
    const int FILE_FLAG_OVERLAPPED      = unchecked((int)0x40000000);
    const int FILE_FLAG_SEQUENTIAL_SCAN = unchecked((int)0x08000000);

    [DllImport("KERNEL32", SetLastError = true, CharSet = CharSet.Auto, BestFitMapping = false)]
    static extern SafeFileHandle    CreateFile(String              fileName,
                                               int                 desiredAccess,
                                               System.IO.FileShare shareMode,
                                               IntPtr              securityAttrs,
                                               System.IO.FileMode  creationDisposition,
                                               int                 flagsAndAttributes,
                                               IntPtr              templateFile);
    #endregion DLLimport
    /// ////////////////////////////////////////////////////////////////////////
    /// object state.
    FileStream stream;              // the file stream created by this open.  
    SafeFileHandle handle;          // the underlying NTFS file handle. 
    /// <summary>
    /// Given a path and mode/sharing/etc flags, open a file in un-buffered mode.
    ///    (i.e. NTFS is told not to cache the file contents).
    /// </summary>
    /// <param name="path">File name</param>
    /// <param name="mode"> System.IO.FileMode </param>
    /// <param name="acc">System.IO.FileAccess: Read | Write | ReadWrite</param>
    /// <param name="share">System.IO.FileShare</param>
    /// <param name="sequential">sequential file access</param>
    /// <param name="async">async file access</param>
    /// <param name="blockSize">block size in bytes</param>
    /// <returns>Unbuffered file stream.</returns>
    public FileStream openStream(string       path,
                                 FileMode     mode,
                                 FileAccess   acc,
                                 FileShare    share,
                                 bool         sequential,
                                 bool         async,
                                 int          blockSize)
    {
        int flags = FILE_FLAG_NO_BUFFERING;     // default to simmple no buffering
        /* Construct the proper 'flags' value to pass to CreateFile() */
        if  (sequential) flags |= FILE_FLAG_SEQUENTIAL_SCAN;
        if  (async)      flags |= FILE_FLAG_OVERLAPPED;
 
        /* Call the Windows CreateFile() API to open the file */
         handle = CreateFile( path,
                              (int)acc,
                              share,
                              IntPtr.Zero,
                              mode,
                              flags,
                              IntPtr.Zero);

        if (!handle.IsInvalid)
        {   /* Wrap the handle in a stream and return it to the caller */
            stream = new FileStream(handle, acc, blockSize, async);
        }
        else                    // if create call failed to get a handle
        {                       // return a null pointer. 
            stream = null;
            handle = null;
        }
        return stream;
    }
    //////////////////////////////////////////////////////////////////////////
    /// <summary>
    /// Closes the stream opened by an earlier openStream() call.
    /// </summary>
    public void closeStream()
    {
        if  (stream != null)
        {
            stream.Close();
            stream = null;
        }
    }
    //////////////////////////////////////////////////////////////////////////
    /// <summary>
    /// Make sure that we're not holding any stream references (i.e. check
    ///   (i.e. check that the client has called closeStream).
    /// </summary>
    ~OpenFileNoBuff()
    {
        if  (stream != null)
        {
            Console.WriteLine("FATAL ERROR: closeStream() not called, aborting program!");
            System.Environment.Exit(-1);
        }
    }
}
