using System;
using System.IO;
namespace IOexamples
{
    /// <summary>
    ///  ------------------------------------------------------
    /// <h1>
    /// WriteSortBenchmarkFile class     File: WriteSortBenchmarkFile.cs
    /// </h1>
    /// Description: Demonstrate different IO write styles
    /// <ul>
    ///     <li>Byte-At-A-Time</li>
    ///     <li>Line-At-A-Time</li>
    ///     <li>Block-At-A Time</li>
    /// </ul>
    /// by writing the sort benchmark file specified at
    ///             http://research.microsoft.com/barc/SortBenchmark/
    ///
    ///  Version: 1.0,  November 2004
    ///
    ///  Author: Peter Kukol, Jim Gray, Microsoft Corporation
    ///        PeterK@microsoft.com, Gray@Microsoft.com
    ///
    ///  (c) Microsoft Corporation. All rights reserved.
    /// 
    ///</summary>
    public class WriteSortBenchmarkFile
    {
        static string printableChars = "abcdefghijklmnopqrstuvwxyz"
                                     + "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                                     + "1234567890"
                                     + @",./<>?;':""[]\{}|`~!@#$%^&*()_+-=";
        int howMany = printableChars.Length;
        static Random random = new Random();
        /// <summary>
        /// *********************************
        /// </summary>
        /// <value>creates SortBenchmark text file containing the designated number of records.</value>
        public void writeSortBenchmarkFile(string fileName, long recordCount, string option)
        {
            int howMany = printableChars.Length;
            const int RecordLength = 98,
                      RecordLengthWithCRLF = RecordLength + 2;
            try
              {                       // catch any file exceptions
                switch (option)         // switch on file access type
                {
                //--------------------------------------------------------------
                // Write the file one byte at a time injecting newline chars every 100 bytes.
                case "Byte-At-A-Time": // using opens and closes the stream
                    using (FileStream fs = new FileStream(fileName, FileMode.OpenOrCreate))
                    {
                        for (long recordNumber = 0; recordNumber < recordCount; recordNumber++)
                        {
                            for (int i = 0; i < RecordLength; i++)
                            {
                                Byte nextByte = (byte)printableChars[random.Next() % howMany];
                                fs.WriteByte(nextByte);
                            }
                            fs.WriteByte((byte)'\n'); fs.WriteByte((byte)'\r');
                        }
                        fs.Flush();
                    }
                    return;
                //--------------------------------------------------------------
                // Write the file one line at a time.
                case "Line-At-A-Time":
                        // using opens and closes the stream
                    using (StreamWriter sw = new StreamWriter(fileName)) {
                        char[] record = new char[RecordLength];
                        Byte[] randomChars = new Byte[RecordLength];
                        for (long recordNumber = 0; recordNumber < recordCount; recordNumber++)
                        {
                            random.NextBytes(randomChars);
                            for (long i = 0; i < RecordLength; i++)
                                record[i] = (char)(printableChars[randomChars[i] % howMany]);
                            sw.WriteLine(record);           // WriteLine() adds cr-lf
                        }
                        sw.Flush();
                    }
                    return;

                //--------------------------------------------------------------
                // Write the file one 100KB block at a time and count the newline chars
                case "Block-At-A-Time":
                default:
                    using (FileStream fs = new FileStream(fileName, FileMode.OpenOrCreate))
                    {
                        int RecordsPerBlock = 1024;
                        int fileIOblockSize = RecordLengthWithCRLF * RecordsPerBlock; // ~100KB
                        Byte[] randomChars = new Byte[RecordLength];
                        Byte[] IObuff = new byte[fileIOblockSize];  // buffer to hold bytes
                        for (long recordNumber = 0; recordNumber < recordCount; recordNumber += RecordsPerBlock)
                        {
                            int recordsToWrite = (int) Math.Min(RecordsPerBlock, recordCount - recordNumber);
                            for (int i = 0; i < recordsToWrite; i++)
                            {
                                random.NextBytes(randomChars);
                                for (int j = 0; j < RecordLength; j++)
                                {
                                    byte next = 0;
                                    next = (byte)printableChars[randomChars[j] % howMany];
                                    IObuff[RecordLengthWithCRLF * i + j] = next;
                                }
                                IObuff[RecordLengthWithCRLF * i + RecordLength] = (byte)'\r';
                                IObuff[RecordLengthWithCRLF * i + RecordLength + 1] = (byte)'\n';
                            }
                            fs.Write(IObuff, 0, recordsToWrite * RecordLengthWithCRLF);
                        }
                        fs.Flush();
                    }
                    return;
                }
            }
        //--------------------------------------------------------------
        // Any exception gets thrown to here. 
        catch (Exception e)
            {
                Console.WriteLine("Error writing file {0}. \n**************\n {1}'", fileName, e);
                throw new FileNotFoundException("Error writing file: " + fileName, e);
            }
        }
    }
}